from typing import Union
import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles

@njit(cache=True)
def _cmo_numba(source: np.ndarray, period: int) -> np.ndarray:
    n = source.shape[0]
    result = np.empty(n, dtype=np.float64)
    # Initialize result with NaN values
    for i in range(n):
        result[i] = np.nan
    
    if n <= 1:
        return result
    
    # Compute the differences manually
    diff = np.empty(n - 1, dtype=np.float64)
    for i in range(n - 1):
        diff[i] = source[i + 1] - source[i]
    
    # Only compute CMO if we have enough diff values
    if diff.shape[0] >= period:
        for i in range(period, n):
            pos_sum = 0.0
            neg_sum = 0.0
            # Calculate sums over the window diff[i-period:i]
            for j in range(i - period, i):
                d = diff[j]
                if d > 0:
                    pos_sum += d
                elif d < 0:
                    neg_sum += -d
            denom = pos_sum + neg_sum
            if denom == 0.0:
                result[i] = 0.0
            else:
                result[i] = 100.0 * (pos_sum - neg_sum) / denom
    return result


def cmo(candles: np.ndarray, period: int = 14, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    CMO - Chande Momentum Oscillator

    :param candles: np.ndarray
    :param period: int - default: 14
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)
    result = _cmo_numba(source, period)
    return result if sequential else result[-1]



import arrow
from jesse.models.ExchangeApiKeys import ExchangeApiKeys
from jesse.models.NotificationApiKeys import NotificationApiKeys


class AppState:
    def __init__(self):
        self.time = arrow.utcnow().int_timestamp * 1000
        self.starting_time = None
        self.ending_time = None
        self.daily_balance = []

        # used as placeholders for detecting open trades metrics
        self.total_open_trades = 0
        self.total_open_pl = 0
        self.total_liquidations = 0

        self.session_id = ''
        self.session_info = {}

        # live only
        self.exchange_api_key = None
        self.notifications_api_key = None

    def set_session_id(self, session_id) -> None:
        self.session_id = session_id

    def set_exchange_api_key(self, exchange_api_key_id: str) -> None:
        if self.exchange_api_key is not None:
            raise ValueError('exchange_api_key has already been set')

        self.exchange_api_key = ExchangeApiKeys.get_or_none(ExchangeApiKeys.id == exchange_api_key_id)

    def set_notifications_api_key(self, notifications_api_key_id: str) -> None:
        if self.notifications_api_key is not None:
            raise ValueError('notifications_api_key has already been set')

        if notifications_api_key_id == '':
            return

        self.notifications_api_key = NotificationApiKeys.get_or_none(NotificationApiKeys.id == notifications_api_key_id)



import peewee
import jesse.helpers as jh
import numpy as np


class Orderbook(peewee.Model):
    id = peewee.UUIDField(primary_key=True)
    # timestamp in milliseconds
    timestamp = peewee.BigIntegerField()
    symbol = peewee.CharField()
    exchange = peewee.CharField()

    data = peewee.BlobField()

    class Meta:
        from jesse.services.db import database

        database = database.db
        indexes = ((('exchange', 'symbol', 'timestamp'), True),)

    def __init__(self, attributes: dict = None, **kwargs) -> None:
        peewee.Model.__init__(self, attributes=attributes, **kwargs)

        if attributes is None:
            attributes = {}

        for a, value in attributes.items():
            setattr(self, a, value)


# # # # # # # # # # # # # # # # # # # # # # # # # # # 
# # # # # # # # # DB FUNCTIONS # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # # # # # # # # 

def store_orderbook_into_db(exchange: str, symbol: str, orderbook: np.ndarray) -> None:
    return
    d = {
        'id': jh.generate_unique_id(),
        'timestamp': jh.now_to_timestamp(),
        'data': orderbook.dumps(),
        'symbol': symbol,
        'exchange': exchange,
    }

    def async_save() -> None:
        Orderbook.insert(**d).on_conflict_ignore().execute()
        print(
            jh.color(
                f'orderbook: {jh.timestamp_to_time(d["timestamp"])}-{exchange}-{symbol}: [{orderbook[0][0][0]}, {orderbook[0][0][1]}], [{orderbook[1][0][0]}, {orderbook[1][0][1]}]',
                'magenta'
            )
        )

    # async call
    threading.Thread(target=async_save).start()


from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles

from .high_pass_2_pole import high_pass_2_pole_fast


def dec_osc(candles: np.ndarray, hp_period: int = 125, k: float = 1, source_type: str = "close",
            sequential: bool = False) -> Union[float, np.ndarray]:
    """
    Ehlers Decycler Oscillator

    :param candles: np.ndarray
    :param hp_period: int - default: 125
    :param k: float - default: 1
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)

    hp = high_pass_2_pole_fast(source, hp_period)
    dec = source - hp
    decosc = high_pass_2_pole_fast(dec, 0.5 * hp_period)

    res = 100 * k * decosc / source

    return res if sequential else res[-1]



from .GateSpotMain import GateSpotMain
from jesse.enums import exchanges


class GateSpot(GateSpotMain):
    def __init__(self) -> None:
        super().__init__(
            name=exchanges.GATE_SPOT,
            rest_endpoint='https://api.gateio.ws/api/v4/spot'
        )



from jesse.enums import timeframes


def timeframe_to_interval(timeframe: str) -> str:
    if timeframe == timeframes.MINUTE_1:
        return '1m'
    elif timeframe == timeframes.MINUTE_5:
        return '5m'
    elif timeframe == timeframes.MINUTE_15:
        return '15m'
    elif timeframe == timeframes.MINUTE_30:
        return '30m'
    elif timeframe == timeframes.HOUR_1:
        return '1h'
    elif timeframe == timeframes.HOUR_3:
        return '3h'
    elif timeframe == timeframes.HOUR_6:
        return '6h'
    elif timeframe == timeframes.HOUR_12:
        return '12h'
    elif timeframe == timeframes.DAY_1:
        return '1D'
    elif timeframe == timeframes.WEEK_1:
        return '1W'
    elif timeframe == timeframes.MONTH_1:
        return '1M'
    else:
        raise NotImplemented('Invalid timeframe: {}'.format(timeframe))


def interval_to_timeframe(interval: str) -> str:
    if interval == '1m':
        return timeframes.MINUTE_1
    elif interval == '5m':
        return timeframes.MINUTE_5
    elif interval == '15m':
        return timeframes.MINUTE_15
    elif interval == '30m':
        return timeframes.MINUTE_30
    elif interval == '1h':
        return timeframes.HOUR_1
    elif interval == '3h':
        return timeframes.HOUR_3
    elif interval == '6h':
        return timeframes.HOUR_6
    elif interval == '12h':
        return timeframes.HOUR_12
    elif interval == '1D':
        return timeframes.DAY_1
    elif interval == '1W':
        return timeframes.WEEK_1
    elif interval == '1M':
        return timeframes.MONTH_1
    else:
        raise NotImplemented('Invalid interval: {}'.format(interval))



from .bollinger_bands import bollinger_bands
from .sma import sma
from .trange import trange
import numpy as np


def ttm_squeeze(candles: np.ndarray, length_ttms: int = 20, bb_mult_ttms: float = 2.0, kc_mult_low_ttms: float = 2.0) -> bool:
    """
    @author daviddtech
    credits: https://www.tradingview.com/script/Mh3EmxF5-TTM-Squeeze-DaviddTech/

    TTMSQUEEZE - TTMSqueeze

    :param candles: np.ndarray
    :param length_ttms: int - default: 20
    :param bb_mult_ttms: float - default: 2.0
    :param kc_mult_low_ttms: float - default: 2.0

    :return: TTMSqueeze(sqz_signal)
    """
    bb_data = bollinger_bands(candles, length_ttms, bb_mult_ttms)

    kc_basis_ttms = sma(candles, length_ttms)
    devkc_ttms = sma(trange(candles, sequential=True), period=length_ttms)

    no_sqz_ttms = bb_data.lowerband < kc_basis_ttms - devkc_ttms * \
        kc_mult_low_ttms or bb_data.upperband > kc_basis_ttms + devkc_ttms * kc_mult_low_ttms

    sqz_signal = False
    if no_sqz_ttms:
        sqz_signal = True

    return sqz_signal



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles

from .high_pass_2_pole import high_pass_2_pole_fast


def decycler(candles: np.ndarray, hp_period: int = 125, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Ehlers Simple Decycler

    :param candles: np.ndarray
    :param hp_period: int - default: 125
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    hp = high_pass_2_pole_fast(source, hp_period)
    res = source - hp

    return res if sequential else res[-1]



from jesse.models.ExchangeApiKeys import ExchangeApiKeys
from jesse.models.NotificationApiKeys import NotificationApiKeys
from jesse.models.OptimizationSession import OptimizationSession
import json
import jesse.helpers as jh

def get_exchange_api_key(exchange_api_key: ExchangeApiKeys) -> dict:
    result = {
        'id': str(exchange_api_key.id),
        'exchange': exchange_api_key.exchange_name,
        'name': exchange_api_key.name,
        'api_key': exchange_api_key.api_key[0:4] + '***...***' + exchange_api_key.api_key[-4:],
        'api_secret': exchange_api_key.api_secret[0:4] + '***...***' + exchange_api_key.api_secret[-4:],
        'created_at': exchange_api_key.created_at.isoformat(),
    }

    if type(exchange_api_key.additional_fields) == str:
        exchange_api_key.additional_fields = json.loads(exchange_api_key.additional_fields)

    # additional fields
    if exchange_api_key.additional_fields:
        for key, value in exchange_api_key.additional_fields.items():
            result[key] = value[0:4] + '***...***' + value[-4:]

    return result


def get_notification_api_key(api_key: NotificationApiKeys, protect_sensitive_data=True) -> dict:
    result = {
        'id': str(api_key.id),
        'name': api_key.name,
        'driver': api_key.driver,
        'created_at': api_key.created_at.isoformat()
    }

    # Parse the fields from the JSON string
    fields = json.loads(api_key.fields)

    # Add each field to the result
    for key, value in fields.items():
        if protect_sensitive_data:
            result[key] = value[0:4] + '***...***' + value[-4:]
        else:
            result[key] = value

    return result


def get_optimization_session(session: OptimizationSession) -> dict:
    """
    Transform an OptimizationSession model instance into a dictionary for API responses
    """
    return {
        'id': str(session.id),
        'status': session.status,
        'completed_trials': session.completed_trials,
        'created_at': session.created_at,
        'updated_at': session.updated_at,
        'best_score': session.best_score,
        'state': json.loads(session.state) if session.state else None
    }


def get_optimization_session_for_load_more(session: OptimizationSession) -> dict:
    objective_function_config = jh.get_config('env.optimization.objective_function', 'sharpe').lower()
    mapping = {
        'sharpe': 'sharpe_ratio',
        'calmar': 'calmar_ratio',
        'sortino': 'sortino_ratio',
        'omega': 'omega_ratio',
        'serenity': 'serenity_index',
        'smart sharpe': 'smart_sharpe',
        'smart sortino': 'smart_sortino'
    }
    metric_key = mapping.get(objective_function_config, objective_function_config)

    best_candidates = []

    def replace_inf_with_null(obj):
        if isinstance(obj, dict):
            return {k: replace_inf_with_null(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [replace_inf_with_null(item) for item in obj]
        elif isinstance(obj, float) and (obj == float('inf') or obj == float('-inf')):
            return None
        return obj

    best_trials_list = replace_inf_with_null(json.loads(session.best_trials)) if session.best_trials else []

    for idx, t in enumerate(best_trials_list):
        training_metrics = t.get('training_metrics', {})
        testing_metrics = t.get('testing_metrics', {})

        train_value = training_metrics.get(metric_key, None)
        test_value = testing_metrics.get(metric_key, None)
        if isinstance(train_value, (int, float)):
            train_value = round(train_value, 2)
        if isinstance(test_value, (int, float)):
            test_value = round(test_value, 2)
        if train_value is None:
            train_value = "N/A"
        if test_value is None:
            test_value = "N/A"

        candidate_objective_metric = f"{train_value} / {test_value}"

        best_candidates.append({
            'rank': f"#{idx + 1}",
            'trial': f"Trial {t['trial']}",
            'params': t['params'],
            'fitness': t['fitness'],
            'dna': t['dna'],
            'training_metrics': training_metrics,  # Use the already fetched metrics
            'testing_metrics': testing_metrics,   # Use the already fetched metrics
            'objective_metric': candidate_objective_metric
        })

    best_candidates = json.dumps(best_candidates).replace('NaN', 'null')
    best_candidates = json.loads(best_candidates)

    objective_curve = None
    if session.objective_curve:
        objective_curve = session.objective_curve.replace('-Infinity', 'null').replace('Infinity', 'null')
        objective_curve = objective_curve.replace('NaN', 'null')
        objective_curve = json.loads(objective_curve)

    return {
        'id': str(session.id),
        'status': session.status,
        'completed_trials': session.completed_trials,
        'total_trials': session.total_trials,
        'created_at': session.created_at,
        'updated_at': session.updated_at,
        'best_score': session.best_score,
        'best_candidates': best_candidates,
        'objective_curve': objective_curve,
        'state': session.state_json,
        'exception': session.exception,
        'traceback': session.traceback
    }



import jesse.helpers as jh
from jesse.services import logger
from jesse.info import exchange_info


def save_daily_portfolio_balance(is_initial=False) -> None:
    if is_initial:
        logger.reset()

    from jesse.store import store

    # # store daily_balance of assets into database
    # if jh.is_livetrading():
    #     for asset_key, asset_value in e.assets.items():
    #         store_daily_balance_into_db({
    #             'id': jh.generate_unique_id(),
    #             'timestamp': jh.now(),
    #             'identifier': jh.get_config('env.identifier', 'main'),
    #             'exchange': e.name,
    #             'asset': asset_key,
    #             'balance': asset_value,
    #         })
    total_balances = 0
    # select the first item in store.exchanges.storage.items()
    try:
        e, = store.exchanges.storage.values()
    except ValueError:
        raise ValueError('Multiple exchange support is not supported at the moment')
    
    if e.type == 'futures':
        # For futures, add wallet balance and sum of all PNLs
        total_balances = e.assets[jh.app_currency()]
        for key, pos in store.positions.storage.items():
            if pos.is_open:
                total_balances += pos.pnl
    else:
        # For spot, just get portfolio_value from any strategy (they all share the same wallet)
        # Get the first strategy we can find
        for key, pos in store.positions.storage.items():
            total_balances = pos.strategy.portfolio_value
            break

    store.app.daily_balance.append(total_balances)

    if not jh.is_livetrading():
        logger.info(f'Saved daily portfolio balance: {round(total_balances, 2)}')


def get_exchange_type(exchange_name: str) -> str:
    """
    a helper for getting the exchange_type for the running session
    """
    # in live trading, exchange type is not configurable, hence we hardcode it
    if jh.is_live():
        return exchange_info[exchange_name]['type']

    # for other trading modes, we can get the exchange type from the config file
    return jh.get_config(f'env.exchanges.{exchange_name}.type')



from collections import namedtuple

from .bollinger_bands import bollinger_bands
from .sma import sma
from .trange import trange
from .linearreg import linearreg
from .stddev import stddev
import numpy as np

SqueezeMomentum = namedtuple('SqueezeMomentum', ['squeeze', 'momentum', 'momentum_signal'])


def squeeze_momentum(candles: np.ndarray, length: int = 20, mult: float = 2.0, length_kc: int = 20, mult_kc: float = 1.5, sequential: bool = True) -> SqueezeMomentum:
    """
    @author lazyBear
    credits: https://www.tradingview.com/script/nqQ1DT5a-Squeeze-Momentum-Indicator-LazyBear/

    squeeze_momentum

    :param candles: np.ndarray
    :length: int - default: 20
    :mult: float - default: 2.0
    :length_kc: float - default: 2.0
    :mult_kc: float - default: 1.5
    :sequential: bool - default: True

    :return: SqueezeMomentum(squeeze, momentum, momentum_signal)
    """
    # calculate bollinger bands
    basis = sma(candles, length, sequential=True)
    dev = mult_kc * stddev(candles, length, sequential=True)
    upper_bb = basis + dev
    lower_bb = basis - dev

    # calculate KC
    ma = sma(candles, length_kc, sequential=True)
    range_ma = sma(trange(candles, sequential=True), period=length_kc, sequential=True)
    upper_kc = ma + range_ma * mult_kc
    lower_kc = ma - range_ma * mult_kc

    sqz = []
    for i in range(len(lower_bb)):
        sqz_on = (lower_bb[i] > lower_kc[i]) and (upper_bb[i] < upper_kc[i])
        sqz_off = (lower_bb[i] < lower_kc[i]) and (upper_bb[i] > upper_kc[i])
        noSqz = (sqz_on == False) and (sqz_off == False)
        sqz.append(0 if noSqz else (-1 if sqz_on else 1))

    highs = np.nan_to_num(_highest(candles[:, 3], length_kc), 0)
    lows = np.nan_to_num(_lowest(candles[:, 4], length_kc), 0)
    sma_arr = np.nan_to_num(sma(candles, period=length_kc, sequential=True))

    momentum = []
    for i in range(len(highs)):
        momentum.append(candles[:, 2][i] - ((highs[i] + lows[i]) / 2 + sma_arr[i]) / 2)

    momentum = linearreg(np.array(momentum), period=length_kc, sequential=True)

    momentum_signal = []
    for i in range(len(momentum) - 1):
        if momentum[i + 1] > 0:
            momentum_signal.append(1 if momentum[i + 1] > momentum[i] else 2)
        else:
            momentum_signal.append(-1 if momentum[i + 1] < momentum[i] else -2)

    if sequential:
        return SqueezeMomentum(sqz, momentum, momentum_signal)
    else:
        return SqueezeMomentum(sqz[-1], momentum[-1], momentum_signal[-1])


def _highest(values, length):
    # Ensure values is a NumPy array for efficient computation
    values = np.asarray(values)
    # Initialize an array to hold the highest values
    highest_values = np.full(values.shape, np.nan)
    # Compute the highest value for each window
    for i in range(length - 1, len(values)):
        highest_values[i] = np.max(values[i - length + 1:i + 1])
    return highest_values


def _lowest(values, length):
    # Ensure values is a NumPy array for efficient computation
    values = np.asarray(values)
    # Initialize an array to hold the lowest values
    lowest_values = np.full(values.shape, np.nan)
    # Compute the lowest value for each window
    for i in range(length - 1, len(values)):
        lowest_values[i] = np.min(values[i - length + 1:i + 1])
    return lowest_values



from .BinanceMain import BinanceMain
from jesse.enums import exchanges


class BinanceUSSpot(BinanceMain):
    def __init__(self) -> None:
        from .BinanceSpot import BinanceSpot

        super().__init__(
            name=exchanges.BINANCE_US_SPOT,
            rest_endpoint='https://api.binance.us/api',
            backup_exchange_class=BinanceSpot
        )



import peewee
from jesse.services.db import database


if database.is_closed():
    database.open_connection()


class Log(peewee.Model):
    id = peewee.UUIDField(primary_key=True)
    session_id = peewee.UUIDField(index=True)
    timestamp = peewee.BigIntegerField()
    message = peewee.TextField()
    # 1: info, 2: error, maybe add more in the future?
    type = peewee.SmallIntegerField()

    class Meta:
        from jesse.services.db import database
        database = database.db
        indexes = (
            (('session_id', 'type', 'timestamp'), False),
        )

    def __init__(self, attributes=None, **kwargs) -> None:
        peewee.Model.__init__(self, attributes=attributes, **kwargs)

        if attributes is None:
            attributes = {}

        for a in attributes:
            setattr(self, a, attributes[a])


# if database is open, create the table
if database.is_open():
    Log.create_table()


# # # # # # # # # # # # # # # # # # # # # # # # # # # 
# # # # # # # # # DB FUNCTIONS # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # # # # # # # # 

def store_log_into_db(log: dict, log_type: str) -> None:
    if log_type == 'info':
        log_type = 1
    elif log_type == 'error':
        log_type = 2
    else:
        raise ValueError(f"Unsupported log_type value: {log_type}")

    d = {
        'id': log['id'],
        'session_id': log['session_id'],
        'type': log_type,
        'timestamp': log['timestamp'],
        'message': log['message']
    }

    Log.insert(**d).execute()



import math
from decimal import Decimal
from typing import Union

import numpy as np
import pandas as pd

import jesse.helpers as jh
from jesse.enums import timeframes
import statsmodels.api as sm

import jesse_rust


def anchor_timeframe(timeframe: str) -> str:
    """
    Returns the anchor timeframe. Useful for writing
    dynamic strategies using multiple timeframes.

    :param timeframe: str
    :return: str
    """

    dic = {
        timeframes.MINUTE_1: timeframes.MINUTE_5,
        timeframes.MINUTE_3: timeframes.MINUTE_15,
        timeframes.MINUTE_5: timeframes.MINUTE_30,
        timeframes.MINUTE_15: timeframes.HOUR_2,
        timeframes.MINUTE_30: timeframes.HOUR_3,
        timeframes.MINUTE_45: timeframes.HOUR_3,
        timeframes.HOUR_1: timeframes.HOUR_4,
        timeframes.HOUR_2: timeframes.HOUR_6,
        timeframes.HOUR_3: timeframes.DAY_1,
        timeframes.HOUR_4: timeframes.DAY_1,
        timeframes.HOUR_6: timeframes.DAY_1,
        timeframes.HOUR_8: timeframes.DAY_1,
        timeframes.HOUR_12: timeframes.DAY_1,
    }

    return dic[timeframe]


def crossed(series1: np.ndarray, series2: Union[float, int, np.ndarray], direction: str = None,
            sequential: bool = False) -> bool:
    """
    Helper for detection of crosses

    :param series1: np.ndarray
    :param series2: float, int, np.array
    :param direction: str - default: None - above or below
    :param sequential: bool - default: False

    :return: bool
    """

    if sequential:
        series1_shifted = jh.np_shift(series1, 1, np.nan)

        if type(series2) is np.ndarray:
            series2_shifted = jh.np_shift(series2, 1, np.nan)
        else:
            series2_shifted = series2

        if direction is None or direction == "above":
            cross_above = np.logical_and(series1 > series2, series1_shifted <= series2_shifted)

        if direction is None or direction == "below":
            cross_below = np.logical_and(series1 < series2, series1_shifted >= series2_shifted)

        if direction is None:
            return np.logical_or(cross_above, cross_below)

    else:
        if type(series2) is not np.ndarray:
            series2 = np.array([series2, series2])

        if direction is None or direction == "above":
            cross_above = series1[-2] <= series2[-2] and series1[-1] > series2[-1]
        if direction is None or direction == "below":
            cross_below = series1[-2] >= series2[-2] and series1[-1] < series2[-1]

        if direction is None:
            return cross_above or cross_below

    if direction == "above":
        return cross_above
    else:
        return cross_below


def estimate_risk(entry_price: float, stop_price: float) -> float:
    """
    estimates the risk per share

    :param entry_price: float
    :param stop_price: float
    :return: float
    """
    if math.isnan(entry_price) or math.isnan(stop_price):
        raise TypeError()

    return abs(entry_price - stop_price)


def limit_stop_loss(entry_price: float, stop_price: float, trade_type: str, max_allowed_risk_percentage: int) -> float:
    """
    Limits the stop-loss price according to the max allowed risk percentage.
    (How many percent you're OK with the price going against your position)

    :param entry_price:
    :param stop_price:
    :param trade_type:
    :param max_allowed_risk_percentage:
    :return: float
    """
    risk = abs(entry_price - stop_price)
    max_allowed_risk = entry_price * (max_allowed_risk_percentage / 100)
    risk = min(risk, max_allowed_risk)
    return (entry_price - risk) if trade_type == 'long' else (entry_price + risk)


def numpy_candles_to_dataframe(candles: np.ndarray, name_date: str = "date", name_open: str = "open",
                               name_high: str = "high",
                               name_low: str = "low", name_close: str = "close",
                               name_volume: str = "volume") -> pd.DataFrame:
    columns = [name_date, name_open, name_close, name_high, name_low, name_volume]
    df = pd.DataFrame(data=candles, index=pd.to_datetime(candles[:, 0], unit="ms"), columns=columns)
    df[name_date] = pd.to_datetime(df.index, unit="ms")
    return df


def qty_to_size(qty: float, price: float) -> float:
    """
    converts quantity to position-size
    example: requesting 2 shares at the price of %50 would return $100

    :param qty: float
    :param price: float
    :return: float
    """
    if math.isnan(qty) or math.isnan(price):
        raise TypeError()

    return qty * price


def risk_to_qty(capital: float, risk_per_capital: float, entry_price: float, stop_loss_price: float, precision: int = 8,
                fee_rate: float = 0) -> float:
    """
    a risk management tool to quickly get the qty based on risk percentage

    :param capital:
    :param risk_per_capital:
    :param entry_price:
    :param stop_loss_price:
    :param precision:
    :param fee_rate:
    :return: float
    """
    risk_per_qty = abs(entry_price - stop_loss_price)
    size = risk_to_size(capital, risk_per_capital, risk_per_qty, entry_price)

    if fee_rate != 0:
        size = size * (1 - fee_rate * 3)

    return size_to_qty(size, entry_price, precision=precision, fee_rate=fee_rate)


def risk_to_size(capital_size: float, risk_percentage: float, risk_per_qty: float, entry_price: float) -> float:
    """
    calculates the size of the position based on the amount of risk percentage you're willing to take
    example: round(risk_to_size(10000, 1, 0.7, 8.6)) == 1229

    :param capital_size:
    :param risk_percentage:
    :param risk_per_qty:
    :param entry_price:
    :return: float
    """
    if risk_per_qty == 0:
        raise ValueError('risk cannot be zero')

    risk_percentage /= 100
    temp_size = ((risk_percentage * capital_size) / risk_per_qty) * entry_price
    return min(temp_size, capital_size)


def size_to_qty(position_size: float, entry_price: float, precision: int = 3, fee_rate: float = 0) -> float:
    """
    converts position-size to quantity
    example: requesting $100 at the entry_price of $50 would return 2
    :param position_size: float
    :param entry_price: float
    :param precision: int
    :param fee_rate:
    :return: float
    """
    # make sure entry_price is not None
    if entry_price is None:
        raise TypeError(f"entry_price is None")

    if math.isnan(position_size) or math.isnan(entry_price):
        raise TypeError(f"position_size: {position_size}, entry_price: {entry_price}")

    if fee_rate != 0:
        position_size *= 1 - fee_rate * 3

    return jh.floor_with_precision(position_size / entry_price, precision)


def subtract_floats(float1: float, float2: float) -> float:
    """
    Subtracts two floats without the rounding issue in Python

    :param float1: float
    :param float2: float

    :return: float
    """
    return jesse_rust.subtract_floats(float1, float2)


def sum_floats(float1: float, float2: float) -> float:
    """
    Sums two floats without the rounding issue in Python

    :param float1: float
    :param float2: float

    :return: float
    """
    return jesse_rust.sum_floats(float1, float2)


def strictly_increasing(series: np.ndarray, lookback: int) -> bool:
    a = series[-lookback:]
    diff = np.diff(a)
    return np.all(diff > 0)


def strictly_decreasing(series: np.ndarray, lookback: int) -> bool:
    a = series[-lookback:]
    diff = np.diff(a)
    return np.all(diff < 0)


def streaks(series: np.ndarray, use_diff=True) -> np.ndarray:
    if use_diff:
        series = np.diff(series)
    pos = np.clip(series, 0, 1).astype(bool).cumsum()
    neg = np.clip(series, -1, 0).astype(bool).cumsum()
    streak = np.where(series >= 0, pos - np.maximum.accumulate(np.where(series <= 0, pos, 0)),
                      -neg + np.maximum.accumulate(np.where(series >= 0, neg, 0)))

    return np.concatenate(
        (np.full((series.shape[0] - streak.shape[0]), np.nan), streak)
    )


def signal_line(series: np.ndarray, period: int = 10, matype: int = 0) -> np.ndarray:
    from jesse.indicators.ma import ma
    return ma(series, period=period, matype=matype, sequential=True)


def kelly_criterion(win_rate: float, ratio_avg_win_loss: float) -> float:
    return win_rate - ((1 - win_rate) / ratio_avg_win_loss)


def prices_to_returns(price_series: np.ndarray) -> np.ndarray:
    """
    converts a series of asset prices to returns.
    """
    pct = np.diff(price_series) / price_series[:-1] * 100
    return jh.same_length(price_series, pct)


def z_score(series: np.ndarray) -> np.ndarray:
    """
    A Z-score is a numerical measurement that describes a value's relationship to the mean of a group of values. Z-score is measured in terms of standard deviations from the mean
    """
    return (series - np.mean(series)) / np.std(series)


def are_cointegrated(price_returns_1: np.ndarray, price_returns_2: np.ndarray, cutoff=0.05) -> bool:
    """
    Uses unit-root test on residuals to test for cointegrated relationship
    See Hamilton (1994) 19.2 for more details
    H_0 is that there is no cointegration i.e. that the residuals have are unit root series (non-stationary)
    We must observe significant p-value to convince ourselves that the series are cointegrated
    """
    from statsmodels.tsa.stattools import coint

    p_value = coint(price_returns_1, price_returns_2)[1]
    return p_value < cutoff


def dd(msg: str) -> None:
    """
    The dd function dumps the given variables and ends execution of the script. 
    Used for debugging. 

    :param msg: str
    """
    print(msg)
    jh.terminate_app()


def combinations_without_repeat(a: np.ndarray, n: int = 2) -> np.ndarray:
    """
    Creates an array containing all combinations of the passed arrays individual values without repetitions. Useful for the optimization mode.
    """
    from itertools import permutations
    if n <= 1:
        raise ValueError("n must be >= 2")
    return np.array(list(permutations(a, n)))


def calculate_alpha_beta(returns1: np.ndarray, returns2: np.ndarray) -> tuple:
    # Add a constant to the independent variable (returns2)
    X = sm.add_constant(returns2)  # Independent variable
    model = sm.OLS(returns1, X).fit()  # Fit the model
    alpha = model.params[0]  # Intercept (alpha)
    beta = model.params[1]  # Slope (beta)
    return alpha, beta


def timeframe_to_one_minutes(timeframe: str) -> int:
    """
    Converts a given timeframe to its equivalent in minutes.

    :param timeframe: str - The timeframe to convert. Supported timeframes include:
        - '1m', '3m', '5m', '15m', '30m', '45m', '1h', '2h', '3h', '4h', '6h', '8h', '12h', 
          '1d', '3d', '1w', '1M'.
    :return: int - The equivalent number of minutes for the given timeframe.

    :raises InvalidTimeframe: If the provided timeframe is not supported.
    """
    from jesse.enums import timeframes
    from jesse.exceptions import InvalidTimeframe

    dic = {
        timeframes.MINUTE_1: 1,
        timeframes.MINUTE_3: 3,
        timeframes.MINUTE_5: 5,
        timeframes.MINUTE_15: 15,
        timeframes.MINUTE_30: 30,
        timeframes.MINUTE_45: 45,
        timeframes.HOUR_1: 60,
        timeframes.HOUR_2: 60 * 2,
        timeframes.HOUR_3: 60 * 3,
        timeframes.HOUR_4: 60 * 4,
        timeframes.HOUR_6: 60 * 6,
        timeframes.HOUR_8: 60 * 8,
        timeframes.HOUR_12: 60 * 12,
        timeframes.DAY_1: 60 * 24,
        timeframes.DAY_3: 60 * 24 * 3,
        timeframes.WEEK_1: 60 * 24 * 7,
        timeframes.MONTH_1: 60 * 24 * 30,
    }

    try:
        return dic[timeframe]
    except KeyError:
        all_timeframes = [timeframe for timeframe in jh.class_iter(timeframes)]
        raise InvalidTimeframe(
            f'Timeframe "{timeframe}" is invalid. Supported timeframes are {", ".join(all_timeframes)}.'
        )



from typing import Union

import numpy as np

from jesse.helpers import slice_candles


def mfi(candles: np.ndarray, period: int = 14, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    MFI - Money Flow Index

    :param candles: np.ndarray
    :param period: int - default: 14
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    # Extract high, low, close and volume from candles array
    high = candles[:, 3]
    low = candles[:, 4]
    close = candles[:, 2]
    volume = candles[:, 5]

    # Compute Typical Price for each candle
    typical_prices = (high + low + close) / 3.0

    # Compute Raw Money Flow = Typical Price * Volume
    raw_mf = typical_prices * volume

    # Initialize positive and negative flows
    pos_flow = np.zeros_like(typical_prices)
    neg_flow = np.zeros_like(typical_prices)
    # For each candle (starting from 1) decide if it's positive or negative money flow
    pos_flow[1:] = np.where(typical_prices[1:] > typical_prices[:-1], raw_mf[1:], 0)
    neg_flow[1:] = np.where(typical_prices[1:] < typical_prices[:-1], raw_mf[1:], 0)

    # Compute rolling sums over the specified period using convolution
    roll_pos = np.convolve(pos_flow, np.ones(period), mode='valid')
    roll_neg = np.convolve(neg_flow, np.ones(period), mode='valid')

    # Compute Money Flow Ratio; handle division by zero by treating as infinity
    with np.errstate(divide='ignore', invalid='ignore'):
        ratio = np.divide(roll_pos, roll_neg, out=np.full_like(roll_pos, np.inf, dtype=float), where=roll_neg != 0)

    # Compute MFI
    mfi_values = 100 - (100 / (1 + ratio))

    # Prepend NaNs for indices that couldn't compute a full period
    pad = np.full(period - 1, np.nan)
    mfi_values = np.concatenate((pad, mfi_values))

    return mfi_values if sequential else mfi_values[-1]



from collections import namedtuple

from .sma import sma
from .ema import ema
from .stddev import stddev

import numpy as np

from jesse.helpers import get_candle_source, slice_candles


def stiffness(candles: np.ndarray, ma_length: int = 100, stiff_length: int = 60, stiff_smooth: int = 3, source_type: str = "close") -> float:
    """
    @author daviddtech
    credits: https://www.tradingview.com/script/MOw6mUQl-Stiffness-Indicator-DaviddTech

    STIFNESS - Stifness

    :param candles: np.ndarray
    :param ma_length: int - default: 100
    :param stiff_length: int - default: 60
    :param stiff_smooth: int - default: 3
    :param source_type: str - default: "close"

    :return: Stiffness(stiffness, threshold)
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, False)
        source = get_candle_source(candles, source_type=source_type)

    bound_stiffness = sma(source, ma_length, sequential=True) - 0.2 * \
        stddev(source, ma_length, sequential=True)
    sum_above_stiffness = _count_price_exceed_series(source, bound_stiffness, stiff_length)
    return ema(np.array(sum_above_stiffness) * 100 / stiff_length, period=stiff_smooth)


def _count_price_exceed_series(close_prices, art_series, length):
    ex_counts = []

    for i in range(len(close_prices)):
        if i < length:
            ex_counts.append(0)
            continue
        count = 0

        for j in range(i - length + 1, i + 1):
            if close_prices[j] > art_series[j]:
                count += 1

        ex_counts.append(count)

    return ex_counts



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles


def edcf(candles: np.ndarray, period: int = 15, source_type: str = "hl2", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Ehlers Distance Coefficient Filter

    :param candles: np.ndarray
    :param period: int - default: 15
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    # Accept normal array too.
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    res = edcf_fast(source, period)

    return res if sequential else res[-1]


@njit(cache=True)
def edcf_fast(source, period):
    newseries = np.full_like(source, np.nan)

    for j in range(2 * period, source.shape[0]):
        num = 0.0
        coefSum = 0.0
        for i in range(period):
            distance = 0.0
            for lb in range(1, period):
                distance += np.power(source[j - i] - source[j - i - lb], 2)
            num += distance * source[j - i]
            coefSum += distance
        newseries[j] = num / coefSum if coefSum != 0 else 0

    return newseries



from typing import Union
import numpy as np
from numba import njit
from jesse.helpers import same_length, slice_candles
from jesse.indicators import sma


@njit
def _emv(high: np.ndarray, low: np.ndarray, volume: np.ndarray, length, div) -> np.ndarray:
    hl2 = (high + low) / 2

    hl2_change = np.zeros_like(hl2)
    hl2_change[1:] = hl2[1:] - hl2[:-1]
    # Calculate EMV
    emv_raw = div * hl2_change * (high - low) / volume

    # Calculate SMA of EMV
    result = np.zeros_like(emv_raw)
    for i in range(length - 1, len(emv_raw)):
        result[i] = np.mean(emv_raw[i - length + 1:i + 1])

    return result


def emv(candles: np.ndarray, length: int = 14, div: int = 10000, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    EMV - Ease of Movement

    :param candles: np.ndarray
    :param length: int - default: 14
    :param div: int - default: 10000
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    high = candles[:, 3]
    low = candles[:, 4]
    volume = candles[:, 5]

    res = _emv(high, low, volume, length, div)

    return same_length(candles, res) if sequential else res[-1]



from .GateUSDTMain import GateUSDTMain
from jesse.enums import exchanges


class GateUSDTPerpetual(GateUSDTMain):
    def __init__(self) -> None:
        super().__init__(
            name=exchanges.GATE_USDT_PERPETUAL,
            rest_endpoint='https://api.gateio.ws/api/v4/futures'
        )



from math import fabs
from typing import Union

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view

from jesse.helpers import get_candle_source, same_length, slice_candles


def fwma(candles: np.ndarray, period: int = 5, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Fibonacci's Weighted Moving Average (FWMA)

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    # Accept normal array too.
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    fibs = fibonacci(n=period)
    swv = sliding_window_view(source, window_shape=period)
    res = np.average(swv, weights=fibs, axis=-1)

    return same_length(candles, res) if sequential else res[-1]


def fibonacci(n: int = 2) -> np.ndarray:
    """Fibonacci Sequence as a numpy array"""
    n = int(fabs(n)) if n >= 0 else 2

    n -= 1
    a, b = 1, 1

    result = np.array([a])

    for _ in range(n):
        a, b = b, a + b
        result = np.append(result, a)

    fib_sum = np.sum(result)
    if fib_sum > 0:
        return result / fib_sum
    else:
        return result



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles


def ema(series: np.ndarray, period: int) -> np.ndarray:
    """Calculates the Exponential Moving Average (EMA) for a series using a simple recursive formula."""
    alpha = 2 / (period + 1)
    out = np.empty_like(series, dtype=float)
    out[0] = series[0]
    for i in range(1, len(series)):
        out[i] = alpha * series[i] + (1 - alpha) * out[i - 1]
    return out


def stoch(series: np.ndarray, period: int) -> np.ndarray:
    """Calculates the stochastic oscillator for a series over the specified period."""
    result = np.full_like(series, np.nan, dtype=float)
    for i in range(len(series)):
        if i < period - 1:
            result[i] = np.nan
        else:
            window = series[i - period + 1: i + 1]
            low = np.min(window)
            high = np.max(window)
            if high == low:
                result[i] = 0
            else:
                result[i] = 100 * ((series[i] - low) / (high - low))
    return result


def stc(candles: np.ndarray, fast_period: int = 23, slow_period: int = 50, k_period: int = 10, d1_period: int = 3, d2_period: int = 3,
        source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    STC - Schaff Trend Cycle (Oscillator)

    The indicator is computed as follows:
    1. macd = EMA(src, fast_period) - EMA(src, slow_period)
    2. k = stoch(macd, k_period) with NaNs replaced by 0
    3. d   = EMA(k, d1_period)  -> First %D
    4. kd  = stoch(d, k_period) with NaNs replaced by 0
    5. stc_val = EMA(kd, d2_period) clamped between 0 and 100

    :param candles: np.ndarray of candles
    :param fast_period: int - default: 23
    :param slow_period: int - default: 50
    :param k_period: int - default: 10, cycle length for stoch computation
    :param d1_period: int - default: 3, used for first EMA smoothing of k
    :param d2_period: int - default: 3, used for EMA smoothing of kd
    :param source_type: str - default: "close"
    :param sequential: bool - default: False. When False, returns only the last value.
    :return: float or np.ndarray depending on sequential
    """
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)

    ema_fast = ema(source, fast_period)
    ema_slow = ema(source, slow_period)
    macd = ema_fast - ema_slow

    k = stoch(macd, k_period)
    k = np.nan_to_num(k, nan=0)

    d_val = ema(k, d1_period)
    kd = stoch(d_val, k_period)
    kd = np.nan_to_num(kd, nan=0)

    stc_val = ema(kd, d2_period)
    stc_val = np.clip(stc_val, 0, 100)

    return stc_val if sequential else stc_val[-1]
    


from typing import Optional
from fastapi import APIRouter, Header
from fastapi.responses import JSONResponse

from jesse.services import auth as authenticator
from jesse.services.web import FeedbackRequestJson, ReportExceptionRequestJson
from jesse.services.multiprocessing import process_manager
import jesse.helpers as jh

router = APIRouter(prefix="/system", tags=["System"])


@router.post("/feedback")
def feedback(json_request: FeedbackRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Send feedback to the Jesse team
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import jesse_trade
    return jesse_trade.feedback(json_request.description, json_request.email)


@router.post("/report-exception")
def report_exception(json_request: ReportExceptionRequestJson,
                     authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Report an exception to the Jesse team
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import jesse_trade
    
    return jesse_trade.report_exception(
        json_request.description,
        json_request.traceback,
        json_request.mode,
        json_request.attach_logs,
        json_request.session_id,
        json_request.email,
        has_live=jh.has_live_trade_plugin()
    )


@router.post("/general-info")
def general_info(authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Get general information about the system
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services.general_info import get_general_info

    try:
        data = get_general_info(has_live=jh.has_live_trade_plugin())
    except Exception as e:
        jh.error(str(e))
        return JSONResponse({
            'error': str(e)
        }, status_code=500)

    return JSONResponse(
        data,
        status_code=200
    )


@router.post("/active-workers")
def active_workers(authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Get a list of active workers
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    return JSONResponse({
        'data': list(process_manager.active_workers)
    }, status_code=200)



from typing import Union

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view

from jesse.helpers import slice_candles


def beta(candles: np.ndarray, benchmark_candles: np.ndarray, period: int = 5, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    BETA - compares the given candles close price to its benchmark (should be in the same time frame)

    :param candles: np.ndarray
    :param benchmark_candles: np.ndarray
    :param period: int - default: 5
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    benchmark_candles = slice_candles(benchmark_candles, sequential)

    x = candles[:, 2]
    y = benchmark_candles[:, 2]

    if len(x) < period:
        out = np.full_like(x, fill_value=np.nan, dtype=float)
        return out if sequential else np.nan

    windows_x = sliding_window_view(x, window_shape=period)
    windows_y = sliding_window_view(y, window_shape=period)

    mean_x = windows_x.mean(axis=1)
    mean_y = windows_y.mean(axis=1)

    diff_x = windows_x - mean_x[:, None]
    diff_y = windows_y - mean_y[:, None]

    numerator = (diff_x * diff_y).sum(axis=1)
    denominator = (diff_y ** 2).sum(axis=1)

    with np.errstate(divide='ignore', invalid='ignore'):
         beta_vals = numerator / denominator

    out = np.full_like(x, fill_value=np.nan, dtype=float)
    out[period - 1:] = beta_vals

    return out if sequential else out[-1]



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles


def smma(candles: np.ndarray, period: int = 5, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    SMMA - Smoothed Moving Average

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    res = numpy_ewma(source, period)

    return res if sequential else res[-1]


def numpy_ewma(data, window):
    """

    :param data:
    :param window:
    :return:
    """
    alpha = 1 / window
    # scale = 1 / (1 - alpha)
    n = data.shape[0]
    scale_arr = (1 - alpha) ** (-1 * np.arange(n))
    weights = (1 - alpha) ** np.arange(n)
    pw0 = (1 - alpha) ** (n - 1)
    mult = data * pw0 * scale_arr
    cumsums = mult.cumsum()
    return cumsums * scale_arr[::-1] / weights.cumsum()



from typing import Union

import numpy as np

from jesse.helpers import slice_candles


def wclprice(candles: np.ndarray, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    WCLPRICE - Weighted Close Price

    :param candles: np.ndarray
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    # Calculate weighted close price as (high + low + 2*close) / 4, with high=candles[:,3], low=candles[:,4], close=candles[:,2]
    res = (candles[:,3] + candles[:,4] + 2 * candles[:,2]) / 4.0

    return res if sequential else res[-1]



import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import jesse.helpers as jh
from jesse.modes.import_candles_mode.drivers.interface import CandleExchange
from typing import Union
from jesse import exceptions
from .bybit_utils import timeframe_to_interval


class BybitMain(CandleExchange):
    def __init__(self, name: str, rest_endpoint: str, category: str) -> None:
        from jesse.modes.import_candles_mode.drivers.Binance.BinanceSpot import BinanceSpot

        super().__init__(name=name, count=200, rate_limit_per_second=10, backup_exchange_class=BinanceSpot)
        self.name = name
        self.endpoint = rest_endpoint
        self.category = category

        # Setup session with retry strategy
        self.session = requests.Session()
        retries = Retry(
            total=3,
            backoff_factor=1,
            status_forcelist=[408, 429, 500, 502, 503, 504],
            allowed_methods=["HEAD", "GET", "POST"]
        )
        self.session.mount('https://', HTTPAdapter(max_retries=retries, pool_maxsize=100))

    def get_starting_time(self, symbol: str) -> int:
        dashless_symbol = jh.dashless_symbol(symbol)
        payload = {
            'category': self.category,
            'symbol': dashless_symbol,
            'interval': 'W',
            'limit': 200,
            'start': 1514811660000
        }

        response = self.session.get(self.endpoint + '/v5/market/kline', params=payload, timeout=10)
        self.validate_response(response)
        data = response.json()['result']['list']
        # Reverse the data list
        data = data[::-1]
        return int(data[1][0])

    def fetch(self, symbol: str, start_timestamp: int, timeframe: str = '1m') -> Union[list, None]:
        dashless_symbol = jh.dashless_symbol(symbol)
        interval = timeframe_to_interval(timeframe)
        payload = {
            'category': self.category,
            'symbol': dashless_symbol,
            'interval': interval,
            'start': int(start_timestamp),
            'limit': self.count
        }

        response = self.session.get(self.endpoint + '/v5/market/kline', params=payload, timeout=10)

        if response.json()['retMsg'] != 'OK':
            raise exceptions.SymbolNotFound(response.json()['retMsg'])
        data = response.json()['result']['list']
        # Reverse the data list
        data = data[::-1]

        return [
            {
                'id': jh.generate_unique_id(),
                'exchange': self.name,
                'symbol': symbol,
                'timeframe': timeframe,
                'timestamp': int(d[0]),
                'open': float(d[1]),
                'close': float(d[4]),
                'high': float(d[2]),
                'low': float(d[3]),
                'volume': float(d[5])
            } for d in data
        ]

    def get_available_symbols(self) -> list:
        response = self.session.get(self.endpoint + '/v5/market/instruments-info?limit=1000&category=' + self.category, timeout=10)
        self.validate_response(response)
        data = response.json()['result']['list']
        return [jh.dashy_symbol(d['symbol']) for d in data]



from typing import Union

import numpy as np

from jesse.helpers import slice_candles


def ad(candles: np.ndarray, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    AD - Chaikin A/D Line

    :param candles: np.ndarray
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    high = candles[:, 3].astype(np.float64)
    low = candles[:, 4].astype(np.float64)
    close = candles[:, 2].astype(np.float64)
    volume = candles[:, 5].astype(np.float64)

    # Calculate Money Flow Multiplier. Safeguard division by zero in case high equals low.
    mfm = np.where(high != low, ((close - low) - (high - close)) / (high - low), 0.0)

    # Calculate Money Flow Volume
    mfv = mfm * volume

    # Compute the Accumulation/Distribution line as the cumulative sum of Money Flow Volume
    ad_line = np.cumsum(mfv)

    return ad_line if sequential else ad_line[-1]



from typing import Union

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view

from jesse.helpers import slice_candles
from jesse.indicators.mean_ad import mean_ad
from jesse.indicators.median_ad import median_ad


def devstop(candles: np.ndarray, period: int = 20, mult: float = 0, devtype: int = 0, direction: str = "long",
            sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Kase Dev Stops

    :param candles: np.ndarray
    :param period: int - default: 20
    :param mult: float - default: 0
    :param devtype: int - default: 0
    :param direction: str - default: long
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    high = candles[:, 3]
    low = candles[:, 4]

    AVTR = rolling_mean(rolling_max(high, 2) - rolling_min(low, 2), period)

    if devtype == 0:
       SD = rolling_std(rolling_max(high, 2) - rolling_min(low, 2), period)
    elif devtype == 1:
       SD = mean_ad(rolling_max(high, 2) - rolling_min(low, 2), period, sequential=True)
    elif devtype == 2:
       SD = median_ad(rolling_max(high, 2) - rolling_min(low, 2), period, sequential=True)

    if direction == "long":
        res = rolling_max(high - AVTR - mult * SD, period)
    else:
        res = rolling_min(low + AVTR + mult * SD, period)

    return res if sequential else res[-1]

def rolling_max(arr, window):
    if len(arr) < window:
        return np.full(arr.shape, np.nan)
    windows = sliding_window_view(arr, window)
    res = np.empty(len(arr))
    res[:window-1] = np.nan
    res[window-1:] = np.max(windows, axis=1)
    return res

def rolling_min(arr, window):
    if len(arr) < window:
        return np.full(arr.shape, np.nan)
    windows = sliding_window_view(arr, window)
    res = np.empty(len(arr))
    res[:window-1] = np.nan
    res[window-1:] = np.min(windows, axis=1)
    return res

def rolling_mean(arr, window):
    if len(arr) < window:
        return np.full(arr.shape, np.nan)
    windows = sliding_window_view(arr, window)
    res = np.empty(len(arr))
    res[:window-1] = np.nan
    res[window-1:] = np.mean(windows, axis=1)
    return res

def rolling_std(arr, window):
    if len(arr) < window:
        return np.full(arr.shape, np.nan)
    windows = sliding_window_view(arr, window)
    res = np.empty(len(arr))
    res[:window-1] = np.nan
    res[window-1:] = np.std(windows, axis=1, ddof=0)
    return res



from jesse.strategies import Strategy


class TestEmptyStrategy(Strategy):
    def should_long(self):
        return False

    def should_short(self):
        return False

    def should_cancel_entry(self):
        return False

    def go_long(self):
        pass

    def go_short(self):
        pass



from typing import Union
import numpy as np
from jesse.helpers import get_candle_source, slice_candles
from jesse_rust import bollinger_bands_width as bollinger_bands_width_rust


def bollinger_bands_width(candles: np.ndarray, period: int = 20, mult: float = 2.0, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    BBW - Bollinger Bands Width

    :param candles: np.ndarray
    :param period: int - default: 20
    :param mult: float - default: 2
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)
    
    result = bollinger_bands_width_rust(source, period, float(mult))
    
    return result if sequential else result[-1]



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles
from jesse.indicators.ma import ma
from jesse.indicators.mean_ad import mean_ad
from jesse.indicators.median_ad import median_ad


def zscore(candles: np.ndarray, period: int = 14, matype: int = 0, nbdev: float = 1, devtype: int = 0, source_type: str = "close",
           sequential: bool = False) -> Union[float, np.ndarray]:
    """
    zScore

    :param candles: np.ndarray
    :param period: int - default: 14
    :param matype: int - default: 0
    :param nbdev: float - default: 1
    :param devtype: int - default: 0
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    if matype == 24 or matype == 29:
        means = ma(candles, period=period, matype=matype, source_type=source_type, sequential=True)
    else:
        means = ma(source, period=period, matype=matype, sequential=True)

    if devtype == 0:
        if len(source) < period:
            sigmas = np.full_like(source, np.nan, dtype=np.float64)
        else:
            # Create a sliding window view of the source array
            rolling_windows = np.lib.stride_tricks.sliding_window_view(source, window_shape=period)
            # Calculate std using population formula (ddof=0)
            std_values = np.std(rolling_windows, axis=1, ddof=0)
            sigmas = np.full(source.shape, np.nan, dtype=np.float64)
            sigmas[period-1:] = std_values
        sigmas = sigmas * nbdev
    elif devtype == 1:
       sigmas = mean_ad(source, period, sequential=True) * nbdev
    elif devtype == 2:
       sigmas = median_ad(source, period, sequential=True) * nbdev

    zScores = (source - means) / sigmas

    return zScores if sequential else zScores[-1]



from jesse.strategies import Strategy
import jesse.helpers as jh


# test_is_smart_enough_to_open_positions_via_market_orders
class Test05(Strategy):
    def update(self):
        pass

    def should_long(self):
        return self.time == 1547201100000 + 60_000

    def should_short(self):
        return self.time == 1547203560000 + 60_000

    def go_long(self):
        qty = 10.204
        self.buy = qty, self.price
        self.stop_loss = qty, 128.35
        self.take_profit = qty, 131.29

    def go_short(self):
        qty = 10
        self.sell = qty, self.price
        self.stop_loss = qty, 129.52
        self.take_profit = qty, 126.58

    def should_cancel_entry(self):
        return False



from .HyperliquidPerpetualMain import HyperliquidPerpetualMain
from jesse.enums import exchanges


class HyperliquidPerpetualTestnet(HyperliquidPerpetualMain):
    def __init__(self) -> None:
        super().__init__(
            name=exchanges.HYPERLIQUID_PERPETUAL_TESTNET,
            rest_endpoint='https://api.hyperliquid-testnet.xyz/info'
        )



import json
from typing import Optional
from starlette.responses import JSONResponse
from jesse.info import live_trading_exchanges
import jesse.helpers as jh
from jesse.services import transformers


def get_exchange_api_keys() -> JSONResponse:
    from jesse.services.db import database
    database.open_connection()

    from jesse.models.ExchangeApiKeys import ExchangeApiKeys

    try:
        # fetch all the api keys
        api_keys = ExchangeApiKeys.select()
    except Exception as e:
        database.close_connection()
        return JSONResponse({
            'status': 'error',
            'message': str(e)
        }, status_code=500)

    # transform each api_key using transformers.get_exchange_api_key()
    api_keys = [transformers.get_exchange_api_key(api_key) for api_key in api_keys]

    database.close_connection()

    return JSONResponse({
        'data': api_keys
    }, status_code=200)


def store_exchange_api_keys(
        exchange: str,
        name: str,
        api_key: str,
        api_secret: str,
        additional_fields: Optional[dict] = None,
        general_notifications_id: Optional[str] = None,
        error_notifications_id: Optional[str] = None,
) -> JSONResponse:
    # validate the exchange
    if exchange not in live_trading_exchanges:
        return JSONResponse({
            'status': 'error',
            'message': f'Invalid exchange: {exchange}'
        }, status_code=400)

    from jesse.services.db import database
    database.open_connection()

    from jesse.models.ExchangeApiKeys import ExchangeApiKeys

    # check if the api key already exists
    if ExchangeApiKeys.select().where(ExchangeApiKeys.name == name).exists():
        database.close_connection()
        return JSONResponse({
            'status': 'error',
            'message': f'API key with the name "{name}" already exists. Please choose another name.'
        }, status_code=400)

    # Ensure additional_fields is a dictionary
    if additional_fields is None:
        additional_fields = {}

    try:
        # create the record
        exchange_api_key: ExchangeApiKeys = ExchangeApiKeys.create(
            id=jh.generate_unique_id(),
            exchange_name=exchange,
            name=name,
            api_key=api_key,
            api_secret=api_secret,
            additional_fields=json.dumps(additional_fields),
            created_at=jh.now_to_datetime(),
            general_notifications_id=general_notifications_id if general_notifications_id else None,
            error_notifications_id=error_notifications_id if error_notifications_id else None
        )
    except ValueError as e:
        database.close_connection()
        return JSONResponse({
            'status': 'error',
            'message': str(e)
        }, status_code=400)
    except Exception as e:
        database.close_connection()
        return JSONResponse({
            'status': 'error',
            'message': str(e)
        }, status_code=500)

    database.close_connection()

    return JSONResponse({
        'status': 'success',
        'message': 'API key has been stored successfully.',
        'data': transformers.get_exchange_api_key(exchange_api_key)
    }, status_code=200)


def delete_exchange_api_keys(exchange_api_key_id: str) -> JSONResponse:
    from jesse.services.db import database
    database.open_connection()

    from jesse.models.ExchangeApiKeys import ExchangeApiKeys

    try:
        # delete the record
        ExchangeApiKeys.delete().where(ExchangeApiKeys.id == exchange_api_key_id).execute()
    except Exception as e:
        database.close_connection()
        return JSONResponse({
            'status': 'error',
            'message': str(e)
        }, status_code=500)

    database.close_connection()

    return JSONResponse({
        'status': 'success',
        'message': 'API key has been deleted successfully.'
    }, status_code=200)



from tabulate import tabulate
from typing import Union

def key_value(data, title: str, alignments: Union[list, tuple] = None, uppercase_title: bool = True) -> None:
    table = [d for d in data]

    if alignments is None:
        print(tabulate(table, headers=[title.upper() if uppercase_title else title, ''], tablefmt="presto"))
    else:
        print(tabulate(table, headers=[title.upper() if uppercase_title else title, ''], tablefmt="presto",
                       colalign=alignments))


def multi_value(data, with_headers: bool = True, alignments: Union[list, tuple] = None) -> None:
    """

    :param data:
    :param with_headers:
    :param alignments:
    :return:
    """
    if data is None:
        return

    headers = data.pop(0) if with_headers else ()

    table = data

    if alignments is None:
        print(tabulate(table, headers=headers, tablefmt="presto"))
    else:
        print(tabulate(table, headers=headers, tablefmt="presto", colalign=alignments))



import threading
import time
from typing import List
import multiprocessing as mp
import traceback
from jesse.services.redis import sync_publish, sync_redis
from jesse.services.failure import terminate_session
import jesse.helpers as jh
from jesse.services.env import ENV_VALUES
import os
import signal

# set multiprocessing process type to spawn
mp.set_start_method('spawn', force=True)


class Process(mp.Process):
    def __init__(self, *args, **kwargs):
        mp.Process.__init__(self, *args, **kwargs)

    def run(self):
        try:
            mp.Process.run(self)
        except Exception as e:
            if type(e).__name__ == 'Termination':
                sync_publish('termination', {})
                jh.terminate_app()
            else:
                sync_publish(
                    'exception',
                    {
                        'error': f'{type(e).__name__}: {e}',
                        'traceback': str(traceback.format_exc()),
                    },
                )

                print('Unhandled exception in the process:')
                print(traceback.format_exc())

                terminate_session()


class ProcessManager:
    def __init__(self):
        self._workers: List[Process] = []
        self._pid_to_client_id_map = {}
        self.client_id_to_pid_to_map = {}
        try:
            port = ENV_VALUES.get('APP_PORT', '9000')
        except:
            port = '9000'
            
        self._active_workers_key = f"{port}|active-processes"
        self._cleanup_thread = threading.Thread(target=self._cleanup_finished_workers, daemon=True)
        self._cleanup_thread.start()

    def _reset(self):
        self._workers = []
        self._pid_to_client_id_map = {}
        self.client_id_to_pid_to_map = {}
        # clear all process status
        sync_redis.delete(self._active_workers_key)

    @staticmethod
    def _prefixed_pid(pid):
        return f"{ENV_VALUES['APP_PORT']}|{pid}"

    @staticmethod
    def _prefixed_client_id(client_id):
        return f"{ENV_VALUES['APP_PORT']}|{client_id}"

    def _add_process(self, client_id):
        sync_redis.sadd(self._active_workers_key, client_id)

    def add_task(self, function, *args):
        client_id = args[0]
        w = Process(target=function, args=args)
        self._workers.append(w)
        w.start()

        self._pid_to_client_id_map[self._prefixed_pid(w.pid)] = self._prefixed_client_id(client_id)
        self.client_id_to_pid_to_map[self._prefixed_client_id(client_id)] = self._prefixed_pid(w.pid)
        self._add_process(client_id)

    def get_client_id(self, pid):
        try:
            client_id: str = self._pid_to_client_id_map[self._prefixed_pid(pid)]
        except KeyError:
            return None
        return jh.string_after_character(client_id, '|')

    def get_pid(self, client_id):
        return self.client_id_to_pid_to_map[self._prefixed_client_id(client_id)]

    def cancel_process(self, client_id):
        sync_redis.srem(self._active_workers_key, client_id)

    def flush(self):
        for w in self._workers:
            try:
                # Try terminate first
                w.terminate()
                # Give it a moment to terminate gracefully
                w.join(timeout=3)
                
                # If still alive, wait a brief moment then force kill
                if w.is_alive():
                    time.sleep(0.5)  # Give terminate a chance to complete
                    os.kill(w.pid, signal.SIGKILL)
                    
                w.close()
            except Exception as e:
                jh.debug(f"Error while terminating process: {str(e)}")
                
        self._reset()

    def _cleanup_finished_workers(self):
        while True:
            try:
                for w in self._workers[:]:  # Create a copy of the list to avoid modification during iteration
                    if not w.is_alive():
                        try:
                            w.join(timeout=1)
                            w.close()
                            self._workers.remove(w)
                        except Exception as e:
                            jh.debug(f"Error during worker cleanup: {str(e)}")
            except Exception as e:
                jh.debug(f"Error in cleanup thread: {str(e)}")
            time.sleep(5)

    @property
    def active_workers(self) -> set:
        """
        Returns the set of all the processes client_id as a list of strings
        """
        return {client_id.decode('utf-8') for client_id in sync_redis.smembers(self._active_workers_key)}


process_manager = ProcessManager()
# flush all processes on startup to avoid any leftover processes
# process_manager.flush()



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles
from jesse.indicators.ma import ma
from jesse.indicators.mean_ad import mean_ad
from jesse.indicators.median_ad import median_ad

def moving_std(source: np.ndarray, window: int) -> np.ndarray:
    n = len(source)
    stdArr = np.empty_like(source)
    if n < window:
        # Vectorized cumulative standard deviation for all indices
        cumsum = np.cumsum(source)
        cumsum2 = np.cumsum(source**2)
        counts = np.arange(1, n + 1)
        means = cumsum / counts
        variances = cumsum2 / counts - means**2
        stdArr[:] = np.sqrt(np.maximum(variances, 0))
    else:
        # For indices with less than a full window, use cumulative statistics
        cumsum_init = np.cumsum(source[:window-1])
        cumsum2_init = np.cumsum(source[:window-1]**2)
        counts_init = np.arange(1, window)
        means_init = cumsum_init / counts_init
        variances_init = cumsum2_init / counts_init - means_init**2
        stdArr[:window-1] = np.sqrt(np.maximum(variances_init, 0))
        
        # For full windows, use sliding window view and compute standard deviation
        sw = np.lib.stride_tricks.sliding_window_view(source, window_shape=window)
        stdArr[window-1:] = np.std(sw, axis=1)
    return stdArr


def vlma(candles: np.ndarray, min_period: int = 5, max_period: int = 50, matype: int = 0, devtype: int = 0, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Variable Length Moving Average

    :param candles: np.ndarray
    :param min_period: int - default: 5
    :param max_period: int - default: 50
    :param matype: int - default: 0
    :param devtype: int - default: 0
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    # Accept normal array too.
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    if matype == 24 or matype == 29:
        mean = ma(candles, period=max_period, matype=matype, source_type=source_type, sequential=True)
    else:
        mean = ma(source, period=max_period, matype=matype, sequential=True)

    if devtype == 0:
       stdDev = moving_std(source, max_period)
    elif devtype == 1:
       stdDev = mean_ad(source, max_period, sequential=True)
    elif devtype == 2:
       stdDev = median_ad(source, max_period, sequential=True)

    a = mean - (1.75 * stdDev)
    b = mean - (0.25 * stdDev)
    c = mean + (0.25 * stdDev)
    d = mean + (1.75 * stdDev)

    res = vlma_fast(source, a, b, c, d , min_period, max_period)

    return res if sequential else res[-1]


@njit(cache=True)
def vlma_fast(source, a, b, c, d, min_period, max_period):
    newseries = np.copy(source)
    period = np.zeros_like(source)
    for i in range(1, source.shape[0]):
        nz_period = period[i - 1] if period[i - 1] != 0 else max_period
        period[i] = nz_period + 1 if b[i] <= source[i] <= c[i] else nz_period - 1 if source[i] < a[i] or source[i] > d[i] else nz_period
        period[i] = max(min(period[i], max_period), min_period)
        sc = 2 / (period[i] + 1)
        newseries[i] = (source[i] * sc) + ((1 - sc) * newseries[i - 1])
    return newseries



from jesse.enums import exchanges
from jesse.modes.import_candles_mode.drivers.Binance.BinanceSpot import BinanceSpot
from jesse.modes.import_candles_mode.drivers.Binance.BinanceUSSpot import BinanceUSSpot
from jesse.modes.import_candles_mode.drivers.Binance.BinancePerpetualFutures import BinancePerpetualFutures
from jesse.modes.import_candles_mode.drivers.Bitfinex.BitfinexSpot import BitfinexSpot
from jesse.modes.import_candles_mode.drivers.Coinbase.CoinbaseSpot import CoinbaseSpot
from jesse.modes.import_candles_mode.drivers.Binance.BinancePerpetualFuturesTestnet import BinancePerpetualFuturesTestnet
from jesse.modes.import_candles_mode.drivers.Bybit.BybitUSDTPerpetual import BybitUSDTPerpetual
from jesse.modes.import_candles_mode.drivers.Bybit.BybitUSDTPerpetualTestnet import BybitUSDTPerpetualTestnet
from jesse.modes.import_candles_mode.drivers.Bybit.BybitUSDCPerpetual import BybitUSDCPerpetual
from jesse.modes.import_candles_mode.drivers.Bybit.BybitUSDCPerpetualTestnet import BybitUSDCPerpetualTestnet
from jesse.modes.import_candles_mode.drivers.Bybit.BybitSpotTestnet import BybitSpotTestnet
from jesse.modes.import_candles_mode.drivers.Bybit.BybitSpot import BybitSpot
from jesse.modes.import_candles_mode.drivers.Apex.ApexProPerpetualTestnet import ApexProPerpetualTestnet
from jesse.modes.import_candles_mode.drivers.Apex.ApexProPerpetual import ApexProPerpetual
from jesse.modes.import_candles_mode.drivers.Apex.ApexOmniPerpetualTestnet import ApexOmniPerpetualTestnet
from jesse.modes.import_candles_mode.drivers.Apex.ApexOmniPerpetual import ApexOmniPerpetual
from jesse.modes.import_candles_mode.drivers.Gate.GateUSDTPerpetual import GateUSDTPerpetual
from jesse.modes.import_candles_mode.drivers.Gate.GateSpot import GateSpot
from jesse.modes.import_candles_mode.drivers.Hyperliquid.HyperliquidPerpetual import HyperliquidPerpetual
from jesse.modes.import_candles_mode.drivers.Hyperliquid.HyperliquidPerpetualTestnet import HyperliquidPerpetualTestnet


drivers = {
    # Perpetual Futures
    exchanges.BINANCE_PERPETUAL_FUTURES: BinancePerpetualFutures,
    exchanges.BINANCE_PERPETUAL_FUTURES_TESTNET: BinancePerpetualFuturesTestnet,
    exchanges.BITFINEX_SPOT: BitfinexSpot,
    exchanges.COINBASE_SPOT: CoinbaseSpot,
    exchanges.BYBIT_USDT_PERPETUAL: BybitUSDTPerpetual,
    exchanges.BYBIT_USDT_PERPETUAL_TESTNET: BybitUSDTPerpetualTestnet,
    exchanges.BYBIT_USDC_PERPETUAL: BybitUSDCPerpetual,
    exchanges.BYBIT_USDC_PERPETUAL_TESTNET: BybitUSDCPerpetualTestnet,
    exchanges.APEX_PRO_PERPETUAL_TESTNET: ApexProPerpetualTestnet,
    exchanges.APEX_PRO_PERPETUAL: ApexProPerpetual,
    exchanges.APEX_OMNI_PERPETUAL_TESTNET: ApexOmniPerpetualTestnet,
    exchanges.APEX_OMNI_PERPETUAL: ApexOmniPerpetual,
    exchanges.GATE_USDT_PERPETUAL: GateUSDTPerpetual,
    exchanges.GATE_SPOT: GateSpot,
    exchanges.HYPERLIQUID_PERPETUAL: HyperliquidPerpetual,
    exchanges.HYPERLIQUID_PERPETUAL_TESTNET: HyperliquidPerpetualTestnet,
    # Spot
    exchanges.BINANCE_SPOT: BinanceSpot,
    exchanges.BINANCE_US_SPOT: BinanceUSSpot,
    exchanges.BYBIT_SPOT_TESTNET: BybitSpotTestnet,
    exchanges.BYBIT_SPOT: BybitSpot,
}


driver_names = list(drivers.keys())



import peewee
import json
from jesse.services.db import database
import jesse.helpers as jh
import json


if database.is_closed():
    database.open_connection()


class OptimizationSession(peewee.Model):
    id = peewee.UUIDField(primary_key=True)
    
    # Status of the optimization session: running, paused, finished, or stopped
    status = peewee.CharField()
    
    # Best trials data in JSON format
    best_trials = peewee.TextField(null=True)
    
    # Objective curve data in JSON format
    objective_curve = peewee.TextField(null=True)
    
    # Frontend state in JSON format - used for restoring UI state
    state = peewee.TextField(null=True)
    
    # Progress tracking
    completed_trials = peewee.IntegerField(default=0)
    total_trials = peewee.IntegerField(default=0)
    exception = peewee.TextField(null=True)
    traceback = peewee.TextField(null=True)

    # Timestamps for session management
    created_at = peewee.BigIntegerField()
    updated_at = peewee.BigIntegerField()

    class Meta:
        from jesse.services.db import database

        database = database.db
        indexes = (
            (('id',), True),
            (('updated_at',), False),
        )

    def __init__(self, attributes: dict = None, **kwargs) -> None:
        peewee.Model.__init__(self, attributes=attributes, **kwargs)

        if attributes is None:
            attributes = {}

        for a, value in attributes.items():
            setattr(self, a, value)
    
    @property
    def best_trials_json(self):
        """
        Returns the best trials as a Python list
        """
        if not self.best_trials:
            return []
        return json.loads(self.best_trials)
    
    @best_trials_json.setter
    def best_trials_json(self, trials_list):
        """
        Sets the best trials from a Python list
        """
        self.best_trials = json.dumps(trials_list)
    
    @property
    def objective_curve_json(self):
        """
        Returns the objective curve data as a Python list
        """
        if not self.objective_curve:
            return []
        return json.loads(self.objective_curve)
    
    @objective_curve_json.setter
    def objective_curve_json(self, curve_data):
        """
        Sets the objective curve data from a Python list
        """
        self.objective_curve = json.dumps(curve_data)
    
    @property
    def state_json(self):
        """
        Returns the frontend state as a Python dictionary
        """
        if not self.state:
            return {}
        return json.loads(self.state)
    
    @state_json.setter
    def state_json(self, state_data):
        """
        Sets the frontend state from a Python dictionary
        """
        self.state = json.dumps(state_data)
    
    @property
    def duration(self):
        """
        Calculate the duration of the session in seconds
        """
        if not self.updated_at:
            # For running sessions, calculate duration up to now
            import jesse.helpers as jh
            return jh.now_to_timestamp(True) - self.created_at
        
        # For completed sessions, use the stored timestamps
        return self.updated_at - self.created_at
    
    @property
    def best_score(self):
        """
        Get the best score from the best trials
        """
        trials = self.best_trials_json
        if not trials:
            return None
            
        # The first trial in the list should be the best one
        # (assuming trials are sorted by score)
        return trials[0].get('fitness', None)


# if database is open, create the table
if database.is_open():
    OptimizationSession.create_table()

# # # # # # # # # # # # # # # # # # # # # # # # # # # 
# # # # # # # # # DB FUNCTIONS # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # # # # # # # # 

def get_optimization_session_by_id(id: str):
    try:
        return OptimizationSession.get(OptimizationSession.id == id)
    except OptimizationSession.DoesNotExist:
        return None


def reset_optimization_session(id: str):
    OptimizationSession.update(
        status='running',
        completed_trials=0,
        best_trials=None,
        objective_curve=None,
        exception=None,
        traceback=None,
        updated_at=jh.now_to_timestamp(True)
    ).where(OptimizationSession.id == id).execute()


def store_optimization_session(
    id: str,
    status: str
) -> None:
    # Create a new session
    d = {
        'id': id,
        'status': status,
        'completed_trials': 0,
        'created_at': jh.now_to_timestamp(True),
        'updated_at': jh.now_to_timestamp(True)
    }
    
    # Save to database
    OptimizationSession.insert(**d).execute()
    

def update_optimization_session_status(id: str, status: str) -> None:
    d = {
        'status': status,
        'updated_at': jh.now_to_timestamp(True)
    }
    
    OptimizationSession.update(**d).where(OptimizationSession.id == id).execute()


def add_session_exception(id: str, exception: str, traceback: str) -> None:
    d = {
        'exception': exception,
        'traceback': traceback,
        'updated_at': jh.now_to_timestamp(True)
    }

    OptimizationSession.update(**d).where(OptimizationSession.id == id).execute()


def update_optimization_session_trials(
    id: str, 
    completed_trials: int, 
    best_trials: list = None,
    objective_curve: list = None,
    total_trials: int = None
) -> None:
    d = {
        'completed_trials': completed_trials,
        'total_trials': total_trials,
        'updated_at': jh.now_to_timestamp(True)
    }

    if best_trials is not None:
        d['best_trials'] = json.dumps(best_trials)

    if objective_curve is not None:
        d['objective_curve'] = json.dumps(objective_curve)

    OptimizationSession.update(**d).where(OptimizationSession.id == id).execute()
    

def get_optimization_session(id: str) -> dict:
    session = OptimizationSession.get(OptimizationSession.id == id)
    return {
        'id': session.id,
        'status': session.status,
        'best_trials': session.best_trials_json,
        'objective_curve': session.objective_curve_json,
        'completed_trials': session.completed_trials,
        'created_at': session.created_at,
        'updated_at': session.updated_at,
        'best_score': session.best_score,
        'state': session.state_json
    }


def get_optimization_sessions() -> list:
    """
    Returns a list of OptimizationSession objects sorted by most recently updated
    """
    return list(OptimizationSession.select().order_by(OptimizationSession.updated_at.desc()))


def delete_optimization_session(id: str) -> bool:
    try:
        OptimizationSession.delete().where(OptimizationSession.id == id).execute()
        return True
    except Exception as e:
        print(f"Error deleting optimization session: {e}")
        return False


def update_optimization_session_state(id: str, state: dict) -> None:
    d = {
        'state': json.dumps(state),
        'updated_at': jh.now_to_timestamp(True)
    }
    
    OptimizationSession.update(**d).where(OptimizationSession.id == id).execute()


from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles

AROON = namedtuple('AROON', ['down', 'up'])


def aroon(candles: np.ndarray, period: int = 14, sequential: bool = False) -> AROON:
    """
    AROON indicator

    :param candles: np.ndarray, expected to have at least 5 columns, with high at index 3 and low at index 4.
    :param period: int - period for the indicator (default: 14)
    :param sequential: bool - whether to return the whole series (default: False)
    :return: AROON(down, up) where each value is computed as above.
    """
    candles = slice_candles(candles, sequential)
    highs = candles[:, 3]
    lows = candles[:, 4]

    if sequential:
        aroon_up = np.full(highs.shape, np.nan, dtype=float)
        aroon_down = np.full(lows.shape, np.nan, dtype=float)
        if len(highs) >= period + 1:
            # Create sliding window view of period+1 elements for highs and lows
            windows_high = np.lib.stride_tricks.sliding_window_view(highs, window_shape=period+1)
            windows_low = np.lib.stride_tricks.sliding_window_view(lows, window_shape=period+1)
            aroon_up[period:] = 100 * (np.argmax(windows_high, axis=1) / period)
            aroon_down[period:] = 100 * (np.argmin(windows_low, axis=1) / period)
        return AROON(down=aroon_down, up=aroon_up)
    else:
        if len(highs) < period + 1:
            up_val = float('nan')
            down_val = float('nan')
        else:
            window_high = highs[-(period + 1):]
            window_low = lows[-(period + 1):]
            up_val = 100 * (np.argmax(window_high) / period)
            down_val = 100 * (np.argmin(window_low) / period)
        return AROON(down=down_val, up=up_val)



import os
import json
import base64
from datetime import timedelta
from multiprocessing import cpu_count
import optuna
import ray
import numpy as np
import jesse.helpers as jh
import jesse.services.logger as logger
from jesse import exceptions
from jesse.services.redis import sync_publish
from jesse.modes.optimize_mode.fitness import get_fitness
from jesse.routes import router
from jesse.services.progressbar import Progressbar
from jesse.services.redis import is_process_active
from jesse.models.OptimizationSession import update_optimization_session_status, update_optimization_session_trials, get_optimization_session, get_optimization_session_by_id
import traceback

# Define a Ray-compatible remote function


@ray.remote
def ray_evaluate_trial(
    user_config,
    formatted_routes,
    formatted_data_routes,
    strategy_hp,
    hp,
    training_warmup_candles,
    training_candles,
    testing_warmup_candles,
    testing_candles,
    optimal_total,
    fast_mode,
    trial_number
):
    """Ray remote function to evaluate a trial"""
    try:
        # Calculate the fitness score using the provided hyperparameters
        score, training_metrics, testing_metrics = get_fitness(
            user_config,
            formatted_routes,
            formatted_data_routes,
            strategy_hp,
            hp,
            training_warmup_candles,
            training_candles,
            testing_warmup_candles,
            testing_candles,
            optimal_total,
            fast_mode
        )

        # Log the trial details if debugging is enabled
        if jh.is_debugging():
            logger.log_optimize_mode(f"Ray Trial {trial_number}: Score={score}, Params={hp}")

        return {
            'trial_number': trial_number,
            'score': score,
            'params': hp,
            'training_metrics': training_metrics,
            'testing_metrics': testing_metrics
        }
    except exceptions.RouteNotFound as e:
        # Convert RouteNotFound to a standard RuntimeError to avoid serialization issues
        error_msg = str(e)
        logger.log_optimize_mode(f"Ray Trial {trial_number} failed with RouteNotFound: {error_msg}")
        logger.log_optimize_mode(f"Trial {trial_number} hyperparameters: {hp}")
        raise RuntimeError(f"RouteNotFound: {error_msg}")
    except Exception as e:
        # Log and re-raise other exceptions
        logger.log_optimize_mode(f"Ray Trial {trial_number} failed with exception: {str(e)}")
        raise

# Optimizer class that uses Ray for hyperparameter optimization


class Optimizer:
    def __init__(
            self,
            session_id: str,
            user_config: dict,
            training_warmup_candles: dict,
            training_candles: dict,
            testing_warmup_candles: dict,
            testing_candles: dict,
            fast_mode: bool,
            optimal_total: int,
            cpu_cores: int,
    ) -> None:
        # Check for Python 3.13 first thing
        if jh.python_version() == (3, 13):
            raise ValueError(
                'Optimization is not supported on Python 3.13. The Ray library used for optimization does not support Python 3.13 yet. Please use Python 3.12 or lower.')

        self.session_id = session_id

        # Retrieve the target strategy and its hyperparameter configuration
        strategy_class = jh.get_strategy_class(router.routes[0].strategy_name)

        self.strategy_hp = strategy_class.hyperparameters(None)

        if not self.strategy_hp:
            update_optimization_session_status(self.session_id, 'stopped')
            raise exceptions.InvalidStrategy('Targeted strategy does not implement a valid hyperparameters() method.')

        # Create study storage for persistence
        os.makedirs('./storage/temp/optuna', exist_ok=True)
        self.storage_url = f"sqlite:///./storage/temp/optuna/optuna_study.db"
        # The study_name uniquely identifies the optimization session - changing it will create a new session
        self.study_name = f"{router.routes[0].strategy_name}_optuna_ray_{self.session_id}"

        self.solution_len = len(self.strategy_hp)
        self.start_time = jh.now_to_timestamp()
        self.fast_mode = fast_mode
        self.optimal_total = optimal_total
        self.training_warmup_candles = training_warmup_candles
        self.training_candles = training_candles
        self.testing_warmup_candles = testing_warmup_candles
        self.testing_candles = testing_candles
        self.user_config = user_config

        # Validate and set the number of CPU cores to use
        if cpu_cores < 1:
            raise ValueError('cpu_cores must be an integer value greater than 0.')
        available = cpu_count()
        self.cpu_cores = cpu_cores if cpu_cores <= available else available

        # Get number of trials from settings
        self.n_trials = self.solution_len * jh.get_config('env.optimization.trials', 200)

        # Create a progress bar instance to update the front end about optimization progress
        self.progressbar = Progressbar(self.n_trials)

        # Initialize best trials tracking
        self.best_trials = []

        # Trial counter and completed trials
        self.trial_counter = 0
        self.completed_trials = 0

        # Create or load the Optuna study for persistence
        self.study = optuna.create_study(
            direction='maximize',
            storage=self.storage_url,
            study_name=self.study_name,
            load_if_exists=True
        )

        # Buffer to accumulate objective curve data points (one point per trial)
        self.objective_curve_buffer = []
        self.total_objective_curve_buffer = []

        # Initialize Ray if not already
        if not ray.is_initialized():
            try:
                ray.init(num_cpus=self.cpu_cores, ignore_reinit_error=True)
                logger.log_optimize_mode(f"Successfully started optimization session with {self.cpu_cores} CPU cores")
            except Exception as e:
                logger.log_optimize_mode(f"Error initializing Ray: {e}. Falling back to 1 CPU.")
                self.cpu_cores = 1
                ray.init(num_cpus=1, ignore_reinit_error=True)

        # Setup a periodic termination check in case the user ends the session
        client_id = jh.get_session_id()
        from timeloop import Timeloop
        self.tl = Timeloop()

        @self.tl.job(interval=timedelta(seconds=1))
        def check_for_termination():
            if is_process_active(client_id) is False:
                # Update session status to 'stopped' in the database
                if get_optimization_session(self.session_id)['status'] != 'terminated':
                    update_optimization_session_status(self.session_id, 'stopped')
                raise exceptions.Termination
        self.tl.start()

        # Load existing trials from the Optuna study
        self._load_study_trials()

    def _load_study_trials(self):
        """Load trials from the database session"""
        session_data = get_optimization_session_by_id(self.session_id)

        def replace_inf_with_null(obj):
            if isinstance(obj, dict):
                return {k: replace_inf_with_null(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [replace_inf_with_null(item) for item in obj]
            elif isinstance(obj, float) and (obj == float('inf') or obj == float('-inf')):
                return None
            return obj
        try:
            # Get completed trials from the Optuna study
            completed_trials = session_data.completed_trials
            if not completed_trials:
                logger.log_optimize_mode("No previous trials found. Starting new optimization session.")
                return

            # Update completed trials count
            self.completed_trials = completed_trials
            logger.log_optimize_mode(f"Loaded {self.completed_trials} completed trials from previous session")
            self.best_trials = replace_inf_with_null(json.loads(session_data.best_trials))
            # Update best candidates display
            self._update_best_candidates()

            # Update progress bar efficiently
            self._set_progressbar_index(self.completed_trials)

            # Set trial counter to start from after the last trial
            self.trial_counter = completed_trials

            self.total_objective_curve_buffer = json.loads(session_data.objective_curve.replace('-Infinity', 'null').replace('Infinity', 'null'))
            # Update the database with loaded trials
            update_optimization_session_trials(
                self.session_id,
                self.completed_trials,
                self.best_trials,
                self.total_objective_curve_buffer,
                self.n_trials
            )

        except Exception as e:
            logger.log_optimize_mode(f"Error loading previous trials: {e}")
            # Reset counters in case of failure
            self.completed_trials = 0
            self.trial_counter = 0

    def _generate_trial_params(self):
        """Generate random hyperparameters for a trial"""
        hp = {}
        for param in self.strategy_hp:
            param_name = str(param['name'])
            param_type = param['type']
            # Convert to string whether input is type class or string
            if isinstance(param_type, type):
                param_type = param_type.__name__
            else:
                # Remove quotes if they exist
                param_type = param_type.strip("'").strip('"')

            if param_type == 'int':
                if 'step' in param and param['step'] is not None:
                    steps = (param['max'] - param['min']) // param['step'] + 1
                    value = param['min'] + np.random.randint(0, steps) * param['step']
                else:
                    value = np.random.randint(param['min'], param['max'] + 1)
                hp[param_name] = value
            elif param_type == 'float':
                if 'step' in param and param['step'] is not None:
                    steps = int((param['max'] - param['min']) / param['step']) + 1
                    value = param['min'] + np.random.randint(0, steps) * param['step']
                else:
                    value = np.random.uniform(param['min'], param['max'])
                hp[param_name] = value
            elif param_type == 'categorical':
                options = param['options']
                hp[param_name] = options[np.random.randint(0, len(options))]
            else:
                raise ValueError(f"Unsupported hyperparameter type: {param_type}")

        return hp

    def _create_optuna_trial(self, trial_number, params, score, training_metrics, testing_metrics):
        """Create and store an Optuna trial for persistence"""
        try:
            # Create distributions for the parameters
            distributions = {}
            for param in self.strategy_hp:
                param_name = str(param['name'])
                param_type = param['type']
                if isinstance(param_type, type):
                    param_type = param_type.__name__
                else:
                    param_type = param_type.strip("'").strip('"')

                if param_type == 'int':
                    if 'step' in param and param['step'] is not None:
                        distributions[param_name] = optuna.distributions.IntDistribution(
                            low=param['min'],
                            high=param['max'],
                            step=param['step']
                        )
                    else:
                        distributions[param_name] = optuna.distributions.IntDistribution(
                            low=param['min'],
                            high=param['max']
                        )
                elif param_type == 'float':
                    if 'step' in param and param['step'] is not None:
                        distributions[param_name] = optuna.distributions.FloatDistribution(
                            low=param['min'],
                            high=param['max'],
                            step=param['step']
                        )
                    else:
                        distributions[param_name] = optuna.distributions.FloatDistribution(
                            low=param['min'],
                            high=param['max']
                        )
                elif param_type == 'categorical':
                    distributions[param_name] = optuna.distributions.CategoricalDistribution(param['options'])

            # Create a new trial
            trial = optuna.create_trial(
                params=params,
                distributions=distributions,
                value=score,
                user_attrs={
                    'training_metrics': training_metrics,
                    'testing_metrics': testing_metrics
                }
            )

            # Add the trial to the study
            self.study.add_trial(trial)
            return True

        except Exception as e:
            logger.log_optimize_mode(f"Error creating Optuna trial: {e}")
            return False

    def _process_trial_result(self, result):
        """Process the result of a completed trial"""
        trial_number = result['trial_number']
        score = result['score']
        params = result['params']
        training_metrics = result['training_metrics']
        testing_metrics = result['testing_metrics']

        # Update progress
        self.completed_trials += 1
        self.progressbar.update()

        # Store trial in Optuna for persistence
        self._create_optuna_trial(trial_number, params, score, training_metrics, testing_metrics)

        # Update the dashboard with general information about the progress
        general_info = {
            'started_at': jh.timestamp_to_arrow(self.start_time).humanize(),
            'trial': f'{self.completed_trials}/{self.n_trials}',
            'objective_function': jh.get_config('env.optimization.objective_function', 'sharpe'),
            'exchange_type': self.user_config['exchange']['type'],
            'leverage_mode': self.user_config['exchange']['futures_leverage_mode'],
            'leverage': self.user_config['exchange']['futures_leverage'],
            'cpu_cores': self.cpu_cores,
        }
        sync_publish('general_info', general_info)

        # Update the progress bar and publish the current progress to the dashboard
        sync_publish('progressbar', {
            'current': self.progressbar.current,
            'estimated_remaining_seconds': self.progressbar.estimated_remaining_seconds
        })

        # Process trial metrics for objective curve
        self._process_trial_metrics(trial_number, training_metrics, testing_metrics)

        # Add to best trials if the score is valid
        if score > 0.0001:
            # Convert parameters to DNA (base64)
            params_str = json.dumps(params, sort_keys=True)
            dna = base64.b64encode(params_str.encode()).decode()

            # Create trial info dict
            current_trial_info = {
                'trial': trial_number,
                'params': params,
                'fitness': round(score, 4),
                'value': score,  # Used for sorting, not sent to frontend
                'dna': dna,
                'training_metrics': training_metrics,
                'testing_metrics': testing_metrics
            }

            # Debug log trial metrics
            if jh.is_debugging():
                jh.debug(f"Trial {trial_number} processed - fitness: {score}")
                jh.debug(f"Trial {trial_number} has training metrics: {bool(training_metrics)}")
                jh.debug(f"Trial {trial_number} has testing metrics: {bool(testing_metrics)}")

            # Insert into best_trials maintaining sorted order
            insert_idx = 0
            for idx, t in enumerate(self.best_trials):
                if score > t['value']:
                    insert_idx = idx
                    break
                else:
                    insert_idx = idx + 1

            if insert_idx < 20 or len(self.best_trials) < 20:
                self.best_trials.insert(insert_idx, current_trial_info)
                # Keep only top 20
                self.best_trials = self.best_trials[:20]

            # Update best candidates table
            self._update_best_candidates()

            # Update the database with the latest trials data
            # We do this every 5 trials to avoid too many database writes
            if self.completed_trials % 5 == 0:

                update_optimization_session_trials(
                    self.session_id,
                    self.completed_trials,
                    self.best_trials,
                    self.total_objective_curve_buffer,
                    self.n_trials
                )

    def _update_best_candidates(self):
        """Update the best candidates table in the dashboard"""
        # Get the objective function configuration
        objective_function_config = jh.get_config('env.optimization.objective_function', 'sharpe').lower()
        mapping = {
            'sharpe': 'sharpe_ratio',
            'calmar': 'calmar_ratio',
            'sortino': 'sortino_ratio',
            'omega': 'omega_ratio',
            'serenity': 'serenity_index',
            'smart sharpe': 'smart_sharpe',
            'smart sortino': 'smart_sortino'
        }
        metric_key = mapping.get(objective_function_config, objective_function_config)

        best_candidates = []
        for idx, t in enumerate(self.best_trials):
            train_value = t.get('training_metrics', {}).get(metric_key, None)
            test_value = t.get('testing_metrics', {}).get(metric_key, None)
            if isinstance(train_value, (int, float)):
                train_value = round(train_value, 2)
            if isinstance(test_value, (int, float)):
                test_value = round(test_value, 2)
            if train_value is None:
                train_value = "N/A"
            if test_value is None:
                test_value = "N/A"

            candidate_objective_metric = f"{train_value} / {test_value}"

            best_candidates.append({
                'rank': f"#{idx + 1}",
                'trial': f"Trial {t['trial']}",
                'params': t['params'],
                'fitness': t['fitness'],
                'dna': t['dna'],
                'training_metrics': t.get('training_metrics', {}),
                'testing_metrics': t.get('testing_metrics', {}),
                'objective_metric': candidate_objective_metric
            })

        # Send top candidates to the dashboard
        sync_publish('best_candidates', best_candidates)

    def _process_trial_metrics(self, trial_number, training_metrics, testing_metrics):
        """Process metrics from a completed trial to update objective curve"""
        # Only add to buffer if both metrics exist and are not empty
        if training_metrics and testing_metrics:
            data_point = {
                'trial': trial_number + 1,
                'training': training_metrics,
                'testing': testing_metrics
            }
            self.objective_curve_buffer.append(data_point)

            if jh.is_debugging():
                jh.debug(f"Added trial {trial_number + 1} to objective curve buffer with metrics")
        else:
            if jh.is_debugging():
                jh.debug(f"Skipped trial {trial_number + 1} - missing metrics. Training: {bool(training_metrics)}, Testing: {bool(testing_metrics)}")

        # Publish a batch every 5 trials or when buffer reaches 10 items
        buffer_size = len(self.objective_curve_buffer)
        if buffer_size >= 10 or (buffer_size > 0 and self.completed_trials % 5 == 0):
            if jh.is_debugging():
                jh.debug(f"Publishing objective curve LEN: {len(self.objective_curve_buffer)}")
            sync_publish('objective_curve', self.objective_curve_buffer)

            if len(self.objective_curve_buffer) > 0 and len(self.total_objective_curve_buffer) > 0:
                if self.objective_curve_buffer[0]['trial'] > self.total_objective_curve_buffer[-1]['trial']:
                    self.total_objective_curve_buffer.extend(self.objective_curve_buffer)
            else:
                self.total_objective_curve_buffer.extend(self.objective_curve_buffer)

            self.objective_curve_buffer = []

    def _set_progressbar_index(self, index):
        """Manually set the progressbar index for resuming sessions efficiently"""
        self.progressbar.index = index
        # Update UI to reflect progress
        sync_publish('progressbar', {
            'current': self.progressbar.current,
            'estimated_remaining_seconds': self.progressbar.estimated_remaining_seconds
        })

    def run(self) -> optuna.trial.FrozenTrial:
        # Log the start of the optimization session
        logger.log_optimize_mode(f"Optimization session started with {self.cpu_cores} CPU cores")

        if self.completed_trials > 0:
            logger.log_optimize_mode(f"Resuming from previous session with {self.completed_trials} trials already completed")
            # Make sure the progress bar is synchronized
            self._set_progressbar_index(self.completed_trials)

        # Track the best trial - handle empty study gracefully
        try:
            best_trial_value = self.study.best_value if self.study.trials else 0.0
            best_trial_params = self.study.best_params if self.study.trials else None
        except (ValueError, AttributeError) as e:
            logger.log_optimize_mode(f"Could not access best trial: {e}. Using default values.")
            best_trial_value = 0.0
            best_trial_params = None

        try:
            # Maximum number of active workers (slightly higher than CPU cores to keep CPUs busy)
            max_workers = min(self.cpu_cores * 2, self.n_trials - self.completed_trials)

            # Dictionary to keep track of active workers
            active_refs = {}
            # Begin optimization loop
            while self.completed_trials < self.n_trials:
                if self.completed_trials == 0:
                    update_optimization_session_trials(
                        self.session_id,
                        0,
                        [],
                        [],
                        self.n_trials
                    )
                # Launch new trials if we have capacity
                while len(active_refs) < max_workers and self.trial_counter < self.n_trials:
                    # Generate parameters for this trial
                    hp = self._generate_trial_params()

                    # Launch the trial evaluation
                    ref = ray_evaluate_trial.options(num_cpus=1).remote(
                        self.user_config,
                        router.formatted_routes,
                        router.formatted_data_routes,
                        self.strategy_hp,
                        hp,
                        self.training_warmup_candles,
                        self.training_candles,
                        self.testing_warmup_candles,
                        self.testing_candles,
                        self.optimal_total,
                        self.fast_mode,
                        self.trial_counter
                    )

                    # Store the reference
                    active_refs[ref] = self.trial_counter
                    self.trial_counter += 1

                # No more workers to launch, wait for results
                if not active_refs:
                    break

                # Wait for any trial to complete (with timeout to ensure responsiveness)
                done_refs, _ = ray.wait(list(active_refs.keys()), num_returns=1, timeout=0.5)

                # Process completed trials
                for ref in done_refs:
                    trial_number = active_refs.pop(ref)
                    try:
                        result = ray.get(ref)
                        # Process the result
                        self._process_trial_result(result)

                        # Update best trial if better
                        if result['score'] > best_trial_value:
                            best_trial_value = result['score']
                            best_trial_params = result['params']
                    except ray.exceptions.RayTaskError as e:
                        # Check if this is a RouteNotFound error converted to RuntimeError
                        if hasattr(e, 'cause') and isinstance(e.cause, RuntimeError) and 'RouteNotFound:' in str(e.cause):
                            raise e.cause
                        else:
                            jh.debug(f'Ray task error for trial {trial_number}: {e}')
                            original_exception = e.cause
                            raise
                    except Exception as e:
                        jh.debug(f'Exception raised in the ray method for trial {trial_number}: {e}')
                        raise e

            # Publish any remaining data in the buffer
            if self.objective_curve_buffer:
                jh.debug(f"Publishing remaining {len(self.objective_curve_buffer)} data points in buffer")
                sync_publish('objective_curve', self.objective_curve_buffer)
                self.objective_curve_buffer = []

            # Get the best trial from the study
            try:
                best_trial = self.study.best_trial
            except (ValueError, AttributeError) as e:
                logger.log_optimize_mode(f"Could not access best trial at the end: {e}")
                # Create a dummy trial if no best trial exists
                best_trial = None

            # Update the database with final results
            update_optimization_session_trials(
                self.session_id,
                self.completed_trials,
                self.best_trials,
                self.total_objective_curve_buffer,
                self.n_trials
            )

            # Update session status to 'finished'
            update_optimization_session_status(self.session_id, 'finished')

            # Publish completion alert
            sync_publish('alert', {
                'message': f"Finished {self.n_trials} trials. Check the \"Best Trials\" table for the best performing parameters.",
                'type': 'success'
            })

        except exceptions.Termination:
            # Handle user-initiated termination
            logger.log_optimize_mode("Optimization terminated by user")
            # Update session status to 'stopped'
            update_optimization_session_status(self.session_id, 'stopped')
            raise
        except Exception as e:
            logger.log_optimize_mode(f"Error during optimization: {e}")
            # Update session status to 'stopped' due to error
            update_optimization_session_status(self.session_id, 'stopped')
            from jesse.models.OptimizationSession import add_session_exception
            add_session_exception(self.session_id, str(e), str(traceback.format_exc()))
            raise
        finally:
            # Shutdown Ray
            ray.shutdown()

        # Create an empty FrozenTrial if best_trial is None
        if best_trial is None:
            logger.log_optimize_mode("No best trial found. Returning empty result.")
            from optuna.trial import FrozenTrial
            return FrozenTrial(
                number=0,
                trial_id=0,
                state=optuna.trial.TrialState.COMPLETE,
                value=0.0,
                datetime_start=None,
                datetime_complete=None,
                params={},
                distributions={},
                user_attrs={},
                system_attrs={},
                intermediate_values={},
            )
        return best_trial



from typing import Union
import numpy as np
from numba import njit
from jesse.helpers import get_candle_source, same_length, slice_candles


@njit
def _wma(arr: np.ndarray, period: int) -> np.ndarray:
    """
    Weighted Moving Average - optimized with Numba
    """
    weights = np.arange(1, period + 1)
    wma = np.zeros_like(arr)

    for i in range(period - 1, len(arr)):
        wma[i] = np.sum(arr[i - period + 1:i + 1] * weights) / np.sum(weights)

    return wma


def hma(candles: np.ndarray, period: int = 5, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    Hull Moving Average

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    # Accept normal array too.
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    # Calculate components for HMA
    half_length = int(period / 2)
    sqrt_length = int(np.sqrt(period))

    # Calculate WMAs
    wma_half = _wma(source, half_length)
    wma_full = _wma(source, period)

    # Calculate 2 * WMA(n/2) - WMA(n)
    raw_hma = 2 * wma_half - wma_full

    # Calculate final HMA
    hma = _wma(raw_hma, sqrt_length)

    return same_length(candles, hma) if sequential else hma[-1]



from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles

# Import the high-performance Rust implementation
from jesse_rust import stoch as stoch_rust  # type: ignore

Stochastic = namedtuple('Stochastic', ['k', 'd'])


def stoch(candles: np.ndarray, fastk_period: int = 14, slowk_period: int = 3, slowk_matype: int = 0,
          slowd_period: int = 3, slowd_matype: int = 0, sequential: bool = False) -> Stochastic:
    """
    The Stochastic Oscillator

    :param candles: np.ndarray
    :param fastk_period: int - default: 14
    :param slowk_period: int - default: 3
    :param slowk_matype: int - default: 0
    :param slowd_period: int - default: 3
    :param slowd_matype: int - default: 0
    :param sequential: bool - default: False

    :return: Stochastic(k, d)
    """
    if any(matype in (24, 29) for matype in (slowk_matype, slowd_matype)):
        raise ValueError("VWMA (matype 24) and VWAP (matype 29) cannot be used in stochastic indicator.")

    candles = slice_candles(candles, sequential)

    # Convert to float64 for Rust compatibility
    candles_f64 = np.asarray(candles, dtype=np.float64)
    
    # Call the Rust implementation
    k, d = stoch_rust(candles_f64, fastk_period, slowk_period, slowk_matype, slowd_period, slowd_matype)
    
    if sequential:
        return Stochastic(k, d)
    else:
        return Stochastic(k[-1], d[-1])



from random import randint
from typing import Union

import numpy as np

# 2021-01-01T00:00:00+00:00
first_timestamp = 1609459080000
open_price = randint(40, 100)
close_price = randint(open_price, 110) if randint(0, 1) else randint(
    30, open_price)
max_price = max(open_price, close_price)
high_price = max_price if randint(0, 1) else randint(max_price, max_price + 10)
min_price = min(open_price, close_price)
low_price = min_price if randint(0, 1) else randint(min_price, min_price + 10)


def range_candles(count: int) -> np.ndarray:
    """
    Generates a range of candles with random values.
    """
    fake_candle(reset=True)
    arr = np.zeros((count, 6))
    for i in range(count):
        arr[i] = fake_candle()
    return arr


def candles_from_close_prices(prices: Union[list, range]) -> np.ndarray:
    """
    Generates a range of candles from a list of close prices.
    The first candle has the timestamp of "2021-01-01T00:00:00+00:00"
    """
    fake_candle(reset=True)
    global first_timestamp
    arr = []
    prev_p = np.nan
    for p in prices:
        # first prev_p
        if np.isnan(prev_p):
            prev_p = p - 0.5

        first_timestamp += 60000
        open_p = prev_p
        close_p = p
        high_p = max(open_p, close_p)
        low_p = min(open_p, close_p)
        vol = randint(0, 200)

        arr.append([first_timestamp, open_p, close_p, high_p, low_p, vol])

        # save prev_p for next candle
        prev_p = p

    return np.array(arr)


def fake_candle(attributes: dict = None, reset: bool = False) -> np.ndarray:
    global first_timestamp
    global open_price
    global close_price
    global max_price
    global high_price
    global min_price
    global low_price

    if reset:
        first_timestamp = 1609459080000
        open_price = randint(40, 100)
        close_price = randint(open_price, 110)
        high_price = max(open_price, close_price)
        low_price = min(open_price, close_price)

    if attributes is None:
        attributes = {}

    first_timestamp += 60000
    open_price = close_price
    close_price += randint(1, 8)
    high_price = max(open_price, close_price)
    low_price = min(open_price - 1, close_price)
    volume = randint(1, 100)
    timestamp = first_timestamp

    return np.array([
        attributes.get('timestamp', timestamp),
        attributes.get('open', open_price),
        attributes.get('close', close_price),
        attributes.get('high', high_price),
        attributes.get('low', low_price),
        attributes.get('volume', volume)
    ], dtype=np.float64)



from fastapi import APIRouter, Header, Query
from typing import Optional, List
from fastapi.responses import JSONResponse, FileResponse
from pydantic import BaseModel

from jesse.services import auth as authenticator
from jesse.services.multiprocessing import process_manager
from jesse.services.web import OptimizationRequestJson, CancelRequestJson, UpdateOptimizationSessionStateRequestJson, UpdateOptimizationSessionStatusRequestJson, TerminateOptimizationRequestJson
from jesse import helpers as jh
from jesse.models.OptimizationSession import get_optimization_sessions as get_sessions, update_optimization_session_state, update_optimization_session_status, delete_optimization_session, reset_optimization_session
from jesse.services.transformers import get_optimization_session, get_optimization_session_for_load_more
from jesse.models.OptimizationSession import get_optimization_session_by_id as get_optimization_session_by_id_from_db
from jesse.modes.optimize_mode import run as run_optimization


router = APIRouter(prefix="/optimization", tags=["Optimization"])


@router.post("")
async def optimization(request_json: OptimizationRequestJson, authorization: Optional[str] = Header(None)):
    """
    Start an optimization process
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    jh.validate_cwd()

    # Check Python version before imports
    if jh.python_version() == (3, 13):
        return JSONResponse({
            'error': 'Optimization is not supported on Python 3.13',
            'message': 'The Ray library used for optimization does not support Python 3.13 yet. Please use Python 3.12 or lower.'
        }, status_code=500)

    process_manager.add_task(
        run_optimization,
        request_json.id,
        request_json.config,
        request_json.exchange,
        request_json.routes,
        request_json.data_routes,
        request_json.training_start_date,
        request_json.training_finish_date,
        request_json.testing_start_date,
        request_json.testing_finish_date,
        request_json.optimal_total,
        request_json.fast_mode,
        request_json.cpu_cores,
        request_json.state,
    )

    return JSONResponse({'message': 'Started optimization...'}, status_code=202)


@router.post("/rerun")
async def rerun_optimization(request_json: OptimizationRequestJson, authorization: Optional[str] = Header(None)):
    """
    Start an optimization process
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    jh.validate_cwd()

    # Check Python version before imports
    if jh.python_version() == (3, 13):
        return JSONResponse({
            'error': 'Optimization is not supported on Python 3.13',
            'message': 'The Ray library used for optimization does not support Python 3.13 yet. Please use Python 3.12 or lower.'
        }, status_code=500)

    # Get the session from the database
    session = get_optimization_session_by_id_from_db(request_json.id)

    if not session:
        return JSONResponse({
            'error': f'Session with ID {request_json.id} not found'
        }, status_code=404)

    # reset the session
    reset_optimization_session(request_json.id)

    process_manager.add_task(
        run_optimization,
        request_json.id,
        request_json.config,
        request_json.exchange,
        request_json.routes,
        request_json.data_routes,
        request_json.training_start_date,
        request_json.training_finish_date,
        request_json.testing_start_date,
        request_json.testing_finish_date,
        request_json.optimal_total,
        request_json.fast_mode,
        request_json.cpu_cores,
        request_json.state,
    )

    return JSONResponse({'message': 'Started optimization...'}, status_code=202)



@router.post("/cancel")
def cancel_optimization(request_json: CancelRequestJson, authorization: Optional[str] = Header(None)):
    """
    Cancel an optimization process
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    process_manager.cancel_process(request_json.id)

    return JSONResponse({'message': f'Optimization process with ID of {request_json.id} was requested for termination'},
                        status_code=202)


@router.get("/download-log")
def download_optimization_log(token: str = Query(...)):
    """
    Download optimization log file
    """
    if not authenticator.is_valid_token(token):
        return authenticator.unauthorized_response()

    from jesse.modes import data_provider

    return data_provider.download_file('optimize', 'log')


@router.post("/sessions")
def get_optimization_sessions(authorization: Optional[str] = Header(None)):
    """
    Get a list of all optimization sessions sorted by most recently updated
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    # Get sessions from the database
    sessions = get_sessions()

    # Transform the sessions using the transformer
    transformed_sessions = [get_optimization_session(session) for session in sessions]

    return JSONResponse({
        'sessions': transformed_sessions,
        'count': len(transformed_sessions)
    })


@router.post("/sessions/{session_id}")
def get_optimization_session_by_id(session_id: str, authorization: Optional[str] = Header(None)):
    """
    Get a single optimization session by ID
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    # Get the session from the database
    session = get_optimization_session_by_id_from_db(session_id)

    if not session:
        return JSONResponse({
            'error': f'Session with ID {session_id} not found'
        }, status_code=404)

    # Transform the session using the transformer
    transformed_session = get_optimization_session_for_load_more(session)

    return JSONResponse({
        'session': transformed_session
    })


@router.post("/update-state")
def update_session_state(request_json: UpdateOptimizationSessionStateRequestJson, authorization: Optional[str] = Header(None)):
    """
    Update the state of an optimization session
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    update_optimization_session_state(request_json.id, request_json.state)

    return JSONResponse({
        'message': 'Optimization session state updated successfully'
    })


@router.post("/terminate")
def terminate_optimization(request_json: TerminateOptimizationRequestJson, authorization: Optional[str] = Header(None)):
    """
    Terminate an optimization process
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    # First update the status to 'terminated'
    update_optimization_session_status(request_json.id, 'terminated')

    # Then request cancellation of the current process
    process_manager.cancel_process(request_json.id)

    return JSONResponse({'message': f'Optimization process with ID of {request_json.id} was terminated'}, status_code=202)


@router.post("/resume")
async def resume_optimization(request_json: OptimizationRequestJson, authorization: Optional[str] = Header(None)):
    """
    Resume an optimization process
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    jh.validate_cwd()

    # Check Python version before imports
    if jh.python_version() == (3, 13):
        return JSONResponse({
            'error': 'Optimization is not supported on Python 3.13',
            'message': 'The Ray library used for optimization does not support Python 3.13 yet. Please use Python 3.12 or lower.'
        }, status_code=500)

    # Get the session from the database
    session = get_optimization_session_by_id_from_db(request_json.id)

    if not session:
        return JSONResponse({
            'error': f'Session with ID {request_json.id} not found'
        }, status_code=404)

    # Transform the session using the transformer
    transformed_session = get_optimization_session_for_load_more(session)

    process_manager.add_task(
        run_optimization,
        request_json.id,
        request_json.config,
        request_json.exchange,
        request_json.routes,
        request_json.data_routes,
        request_json.training_start_date,
        request_json.training_finish_date,
        request_json.testing_start_date,
        request_json.testing_finish_date,
        request_json.optimal_total,
        request_json.fast_mode,
        request_json.cpu_cores,
        request_json.state,
    )

    return JSONResponse({
        'session': transformed_session
    })


@router.post("/sessions/{session_id}/remove")
def remove_optimization_session(session_id: str, authorization: Optional[str] = Header(None)):
    """
    Remove an optimization session from the database
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    session = get_optimization_session_by_id_from_db(session_id)

    if not session:
        return JSONResponse({
            'error': f'Session with ID {session_id} not found'
        }, status_code=404)

    # Delete the session from the database
    result = delete_optimization_session(session_id)

    if not result:
        return JSONResponse({
            'error': f'Failed to delete session with ID {session_id}'
        }, status_code=500)

    return JSONResponse({
        'message': 'Optimization session removed successfully'
    })



from fastapi import APIRouter, Query

from jesse.services import auth as authenticator
from jesse.modes import data_provider


router = APIRouter(prefix="/download", tags=["Download"])


@router.get("/{mode}/{file_type}/{session_id}")
def download(mode: str, file_type: str, session_id: str, token: str = Query(...)):
    """
    Download files such as logs or other generated files.
    Log files require session_id because there is one log per each session. Except for the optimize mode.
    """
    if not authenticator.is_valid_token(token):
        return authenticator.unauthorized_response()

    return data_provider.download_file(mode, file_type, session_id)
